"""
"""
# Standard Library

# Third-Party
from ply import lex, yacc
from cstream import stderr

# Local
from ..satlib import Source, Stack, PythonShell, PythonError, Timing
from ..error import (
    EXIT_FAILURE,
    SatParserError,
    SatSyntaxError,
    SatPythonError,
    SatExit,
)
from ..types import Expr, Var, Number, String, SatType, PythonObject
from ..symbols import SYS_CONFIG, DEF_CONSTANT, DEF_ARRAY, DEF_CONSTRAINT
from ..symbols import T_IDX, T_ADD, T_NEG


def regex(pattern: str):
    def decor(callback):
        callback.__doc__ = pattern
        return callback

    return decor


class SatLexer(object):

    ## List of token names.
    tokens = (
        "FORALL",
        "EXISTS",
        "UNIQUE",  # quantifiers
        "NOT",
        "AND",
        "OR",
        "XOR",  # logical
        "IMP",
        "RIMP",
        "IFF",  # extended logical
        "CONFIG",
        "ENDL",
        "COMMA",
        "ASSIGN",
        "EQ",  # equal
        "GT",  # greater than
        "LT",  # less than
        "GE",  # greater or equal
        "LE",  # less or equal
        "NE",  # not equal
        "MUL",
        "DIV",
        "MOD",
        "ADD",
        "SUB",
        "POW",
        "DOTS",
        "LBRA",
        "RBRA",
        "LPAR",
        "RPAR",
        "LCUR",
        "RCUR",
        "NAME",
        "NUMBER",
        "STRING",
        "PYTHON",
    )

    def __init__(self, source, **kwargs):
        self.lexer = lex.lex(object=self, **kwargs)
        self.source = source

        self.error_stack = Stack()

        ## Separate python shell
        self.python_shell = PythonShell()

    def __lshift__(self, error: SatParserError):
        self.error_stack.push(error)

    def exit(self, code: int):
        raise SatExit(code)

    def interrupt(self):
        """"""
        while self.error_stack:
            error = self.error_stack.popleft()
            stderr[0] << error
        
        self.exit(EXIT_FAILURE)

    def checkpoint(self):
        """"""
        if self.error_stack:
            self.interrupt()

    reserved = {
        "forall": "FORALL",
        "exists": "EXISTS",
        "unique": "UNIQUE",
        "wta": "UNIQUE",
        "sum": "EXISTS",
        "or": "OR",
        "xor": "XOR",
        "and": "AND",
        "iff": "IFF",
        "imp": "IMP",
        "not": "NOT",
        "pow": "POW",
    }

    op_map = {
        "FORALL": "@",
        "EXISTS": "$",
        "UNIQUE": "$!",
        "OR": "|",
        "XOR": "^",
        "AND": "&",
        "IFF": "<->",
        "IMP": "->",
        "NOT": "~",
        "POW": "**"
    }

    # Regular expression rules for tokens
    t_DOTS = r"\:"

    t_LBRA = r"\["
    t_RBRA = r"\]"

    t_LPAR = r"\("
    t_RPAR = r"\)"

    t_LCUR = r"\{"
    t_RCUR = r"\}"

    t_FORALL = r"\@"
    t_EXISTS = r"\$"
    t_UNIQUE = r"\$\!"

    t_NOT = r"\~"

    t_AND = r"\&"

    t_OR = r"\|"

    t_XOR = r"\^"

    t_IMP = r"\-\>"

    t_RIMP = r"\<\-"

    t_IFF = r"\<\-\>"

    t_COMMA = r"\,"

    t_DIV = r"\/"

    t_MOD = r"\%"

    t_POW = r"\*\*"
    t_MUL = r"\*"

    t_ADD = r"\+"

    t_SUB = r"\-"

    t_ENDL = r"\;"

    t_ASSIGN = r"\="

    t_EQ = r"\=\="
    t_GT = r"\>"
    t_LT = r"\<"

    t_GE = r"\>\="
    t_LE = r"\<\="

    t_NE = r"\!\="

    t_CONFIG = r"\?"

    @regex(r"\"(\\.|[^\"])*\"")
    def t_STRING(self, t):
        setattr(t, "string", t.value)
        t.value = String(t.value[1:-1])
        return t

    @regex(r"\`(\\.|[^\`])*\`")
    def t_PYTHON(self, t):
        setattr(t, "string", t.value)
        t.value = str(t.value[1:-1])
        return t

    @regex(r"\n")
    def t_newline(self, _t):
        self.lexer.lineno += 1
        return None

    @regex(r"[a-zA-Z_][a-zA-Z0-9_]*")
    def t_NAME(self, t):
        setattr(t, "string", t.value)
        if t.value in self.reserved:
            t.type = self.reserved[t.value]
            t.value = self.op_map[t.type]
        else:
            t.value = String(t.value)
        t.value = String(t.value)
        return t

    @regex(r"[0-9]*\.?[0-9]+([Ee][-+]?[0-9]+)?")
    def t_NUMBER(self, t):
        setattr(t, "string", t.value)
        t.value = Number(t.value, source=self.source, lexpos=t.lexpos)
        return t

    # String containing ignored characters between tokens
    t_ignore = " \t"
    t_ignore_COMMENT = r"\#.*"

    @regex(r"\#\{[\s\S]*?\}\#")
    def t_ignore_MULTICOMMENT(self, t):
        self.lexer.lineno += str(t.value).count("\n")
        return None

    def t_error(self, t):
        stderr << f"Unknown token '{t.value}' at line {t.lineno}"
        if t:
            stderr << f"Syntax Error at line {t.lineno}:"
            stderr << self.source.lines[t.lineno - 1]
            stderr << f'{" " * (self.chrpos(t.lineno, t.lexpos))}^'
        else:
            stderr << "Unexpected End Of File."

    def chrpos(self, lineno, lexpos):
        return lexpos - self.source.table[lineno - 1] + 1


class SatParser(object):

    tokens = SatLexer.tokens

    precedence = (
        ("left", "ADD", "SUB"),
        ("left", "MUL", "DIV", "MOD"),
        ("left", "POW"),
        ("left", "IMP", "RIMP", "IFF"),
        ("left", "XOR", "OR"),
        ("left", "AND"),
        ("left", "NOT"),
        ("left", "EQ", "NE", "GE", "LE", "GT", "LT"),
        ("left", "FORALL", "EXISTS", "UNIQUE"),
        ("left", "DOTS"),
        ("left", "LBRA", "RBRA"),
        ("left", "LCUR", "RCUR"),
        ("left", "LPAR", "RPAR"),
        ("left", "CONFIG"),
        ("left", "NAME"),
        ("left", "STRING", "NUMBER"),
        ("left", "ENDL"),
        ("left", "PYTHON"),
    )

    def __init__(self):
        ## Input
        self.source = None
        ## Output
        self.bytecode = None

        ## Lex & Yacc
        self.lexer = None
        self.parser = None

        self.error_stack = Stack()

        ## Python Shell
        self.python_shell = PythonShell()

    def __lshift__(self, error: SatParserError):
        self.error_stack.push(error)

    def exit(self, code: int):
        raise SatExit(code)

    def interrupt(self):
        """"""
        while self.error_stack:
            stderr[0] << self.error_stack.popleft()

        self.exit(EXIT_FAILURE)

    def checkpoint(self):
        """"""
        if self.error_stack:
            self.interrupt()

    @Timing.timer(level=2, section="Parser.parse")
    def parse(self, source: Source):
        ## Input
        self.source = source

        ## Initialize Lexer
        self.lexer = SatLexer(self.source)

        ## Initialize Parser
        self.parser = yacc.yacc(module=self, debug=False)

        ## Run Parser
        if not self.source:
            self << SatSyntaxError("Empty File.", target=self.source.eof)
        else:
            self.parser.parse(self.source)

        ## Checkpoint
        self.checkpoint()

        ## Output
        return self.bytecode

    def run(self, code: list):
        self.bytecode = [cmd for cmd in code if cmd is not None]

    def cast_type(self, x: object, type_caster: type):
        """Casts type according to `type_caster(x: object) -> y: object`
        copying `lexinfo` data, if available.
        """
        if hasattr(x, "lexinfo"):
            lexinfo = getattr(x, "lexinfo")
            y = type_caster(x)
            setattr(y, "lexinfo", lexinfo)
            return y
        else:
            return type_caster(x)

    def p_start(self, p):
        """start : code"""
        self.run(p[1])

    def p_code(self, p):
        """code : code stmt
        | stmt
        """
        if len(p) == 3:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_stmt(self, p):
        """stmt : sys_config ENDL
        | def_constant ENDL
        | def_array ENDL
        | def_constraint ENDL
        | cmd_python ENDL
        """
        p[0] = p[1]

    def p_cmd_python(self, p):
        """cmd_python : PYTHON"""
        ## Executes python command
        self.python_shell(p[1], evaluate=False)

        p[0] = None

    def p_sys_config(self, p):
        """sys_config : CONFIG NAME DOTS sys_config_args"""
        name = p[2]
        args = p[4]
        p[0] = (SYS_CONFIG, name, args)

    def p_sys_config_args(self, p):
        """sys_config_args : sys_config_args COMMA sys_config_arg
        | sys_config_arg
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [p[1]]

    def p_sys_config_arg(self, p):
        """sys_config_arg : NUMBER
        | STRING
        """
        p[0] = p[1]

    def p_def_constant(self, p):
        """def_constant : varname ASSIGN expr"""
        p[0] = (DEF_CONSTANT, p[1], p[3])

    def p_literal(self, p):
        """literal : constant
        | python_literal
        | varname
        """
        p[0] = p[1]

    def p_constant(self, p):
        """constant : NUMBER"""
        p[0] = p[1]

    def p_varname(self, p):
        """varname : NAME"""
        p[0] = Var(p[1], source=self.source, lexpos=p.lexpos(1))

    def p_python_literal(self, p):
        """python_literal : PYTHON"""
        py_code = p[1]

        try:
            py_object = PythonObject(self.python_shell(str(py_code), evaluate=True))
        except PythonError as error:
            self << SatPythonError(*error.args)
        finally:
            self.checkpoint()

        self.source.propagate(py_code, py_object)

        p[0] = py_object

    def p_def_array(self, p):
        """def_array : varname shape ASSIGN array_buffer
        | varname shape
        """
        name = p[1]
        shape = p[2]

        if len(p) == 5:  # array declared
            buffer = p[4]
        else:  # implicit array
            buffer = None

        p[0] = (DEF_ARRAY, name, shape, buffer)

    def p_shape(self, p):
        """shape : shape index
        | index
        """
        if len(p) == 3:
            p[0] = (*p[1], p[2])
        else:
            p[0] = (p[1],)

    def p_index(self, p):
        """index : LBRA expr RBRA"""
        p[0] = p[2]

    def p_array_buffer(self, p):
        """array_buffer : LCUR array RCUR
        | python_literal
        """
        if len(p) == 4:  ## default buffer
            p[0] = p[2]
        else:  ## python literal
            p[1]

    def p_array(self, p):
        """array : array COMMA array_item
        | array_item
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [
                p[1],
            ]

    def p_array_item(self, p):
        """array_item : array_index DOTS expr"""
        p[0] = (p[1], p[3])

    def p_array_index(self, p):
        """array_index : LPAR literal_seq RPAR"""
        p[0] = p[2]

    def p_literal_seq(self, p):
        """literal_seq : literal_seq COMMA literal
        | literal
        """
        if len(p) == 4:
            p[0] = (*p[1], p[3])
        else:
            p[0] = (p[1],)

    def p_def_constraint(self, p):
        """def_constraint : LPAR NAME RPAR varname LBRA literal RBRA DOTS loops expr
        | LPAR NAME RPAR varname DOTS loops expr
        """
        type_ = String(p[2], source=self.source, lexpos=p.lexpos(2))
        name = p[4]

        if len(p) == 11:
            level = p[6]
            loops = p[9]
            expr = p[10]
        else:
            level = None
            loops = p[6]
            expr = p[7]

        p[0] = (DEF_CONSTRAINT, type_, name, loops, expr, level)

    def p_loops(self, p):
        """loops : loop_stack
        |
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = []

    def p_loop_stack(self, p):
        """loop_stack : loop_stack loop
        | loop
        """
        if len(p) == 3:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_loop(self, p):
        """loop : quant LCUR varname ASSIGN domain COMMA conditions RCUR
        | quant LCUR varname ASSIGN domain RCUR
        """
        if len(p) == 9:
            p[0] = (p[1], p[3], p[5], p[7])
        else:
            p[0] = (p[1], p[3], p[5], None)

    def p_quant(self, p):
        """quant : FORALL
        | EXISTS
        | UNIQUE
        """
        p[0] = p[1]

    def p_domain(self, p):
        """domain : LBRA expr DOTS expr DOTS expr RBRA
        | LBRA expr DOTS expr RBRA
        """
        if len(p) == 8:
            p[0] = (p[2], p[4], p[6])
        else:
            p[0] = (p[2], p[4], None)

    def p_conditions(self, p):
        """conditions : conditions COMMA condition
        | condition
        """
        if len(p) == 4:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_condition_expr(self, p):
        """condition : expr"""
        p[0] = p[1]

    def p_condition(self, p):
        """expr : expr EQ expr
        | expr GT expr
        | expr LT expr
        | expr GE expr
        | expr LE expr
        | expr NE expr
        """
        p[0] = Expr(p[2], p[1], p[3], source=self.source, lexpos=p.lexpos(2))

    def p_expr(self, p):
        """expr : literal"""
        p[0] = p[1]

    def p_expr1(self, p):
        """expr : NOT expr
        | ADD expr
        | SUB expr
        """
        p[0] = Expr(p[1], p[2], source=self.source, lexpos=p.lexpos(1))

    def p_expr2(self, p):
        """expr : expr AND expr
        | expr OR expr
        | expr XOR expr
        | expr ADD expr
        | expr SUB expr
        | expr MUL expr
        | expr POW expr
        | expr DIV expr
        | expr MOD expr
        | expr IMP expr
        | expr RIMP expr
        | expr IFF expr
        """
        if p[2] == T_NEG:
            p[0] = Expr(
                T_ADD,
                p[1],
                Expr(T_NEG, p[3], source=self.source, lexpos=p.lexpos(2)),
                source=self.source,
                lexpos=p.lexpos(2),
            )
        else:
            p[0] = Expr(p[2], p[1], p[3], source=self.source, lexpos=p.lexpos(2))

    def p_expr_index(self, p):
        """expr : expr LBRA expr RBRA"""
        p[0] = Expr(T_IDX, p[1], p[3], source=self.source, lexpos=p.lexpos(2))

    def p_expr_par(self, p):
        """expr : LPAR expr RPAR"""
        p[0] = p[2]

    def p_error(self, t):
        target = SatType()
        if t:
            target.lexinfo = self.source.getlex(t.lexpos)
            if hasattr(t, "string"):
                msg = f"Unexpected Token '{t.string}'"
            else:
                msg = f"Unexpected Token '{t.value}'"
        else:
            target = self.source.eof
            msg = "Unexpected End Of File"
        self << SatSyntaxError(msg=msg, target=target)
        return None

    def chrpos(self, lineno: int, lexpos: int):
        return lexpos - self.source.table[lineno - 1] + 1



#! /usr/bin/env python3
"""
"""
# Standard Library
import sys

def main(argv: list):
    """
    Parameters
    ----------
    argv: list
        Command line arguments list.
    """
    try:
        from satyrus import SatCLI
    except ImportError:
        raise SystemError("Satyrus3 is not correctly installed. See 'https://satyrus3.github.io/docs/' for installation details.")
    else:
        SatCLI.run(argv)

if __name__ == '__main__':
    main(sys.argv[1:]) # Here we go!


"""
"""

from ..compiler import SatCompiler
from ...types import Number
from ...symbols import PREC, OPT, EPSILON, ALPHA


def run_init(compiler: SatCompiler, *args: tuple):
    """\
    Initializes compiler environment
    """
    default_env = [
        (PREC, 16),
        (OPT, 0),
        (ALPHA, Number("1.0")),
        (EPSILON, Number("1E-4")),
    ]

    for key, val in default_env:
        if key not in compiler.env:
            compiler.env[key] = val



## Standard Library
from collections import deque

class Queue:

    def __init__(self, buffer=None, limit=None):
        self.__limit = limit
        self.__queue = deque([] if buffer is None else buffer, self.__limit)

    def clear(self):
        self.__queue.clear()

    def push(self, x: object):
        self.__queue.appendleft(x)

    def pop(self) -> object:
        return self.__queue.pop()

    def pushleft(self, x: object):
        self.__queue.appendleft(x)

    def popleft(self):
        return self.__queue.popleft()

    def copy(self):
        return self.__class__(buffer=self.__queue.copy(), limit=self.__limit)

    def __getitem__(self, index):
        return self.__queue.__getitem__(index)

    def __str__(self):
        return "\n".join(map(str, self.__queue))

    def __iter__(self):
        return iter(self.__queue)

    def __bool__(self):
        return bool(self.__queue)

    def __len__(self):
        return len(self.__queue)

    def __contains__(self, x):
        return x in self.__queue

    @property
    def top(self):
        return self.__queue[-1]

class Stack:

    def __init__(self, buffer=None, limit=None):
        self.__limit = limit
        self.__stack = deque([] if buffer is None else buffer, self.__limit)

    def clear(self):
        self.__stack.clear()

    def push(self, x: object):
        self.__stack.append(x)

    def pop(self) -> object:
        return self.__stack.pop()

    def pushleft(self, x: object):
        self.__stack.appendleft(x)

    def popleft(self):
        return self.__stack.popleft()

    def copy(self):
        return self.__class__(buffer=self.__stack.copy(), limit=self.__limit)

    def __getitem__(self, index):
        return self.__stack.__getitem__(index)

    def __str__(self):
        return "\n".join(map(str, reversed(self.__stack)))

    def __iter__(self):
        return iter(self.__stack)

    def __bool__(self):
        return bool(self.__stack)

    def __len__(self):
        return len(self.__stack)

    def __contains__(self, x):
        return x in self.__stack

    @property
    def top(self):
        return self.__stack[-1]

__all__ = ["Stack", "Queue"]


"""
"""
# Future Imports
from __future__ import annotations

## Standard Library
from functools import reduce
from typing import Callable

## Third-Party
from cstream import stdlog

## Local
from ..compiler import SatCompiler
from ...satlib import arange, Stack, Posiform
from ...error import (
    SatIndexError,
    SatValueError,
    SatTypeError,
    SatExprError,
    SatWarning,
)
from ...symbols import CONS_INT, CONS_OPT
from ...symbols import T_EXISTS, T_UNIQUE, T_FORALL
from ...symbols import T_AND, T_OR, T_NOT, T_NE, T_ADD, T_NEG, T_MUL, T_XOR, T_POW
from ...types import Expr, Var, String, Number, SatType

LOOP_TYPES = {T_EXISTS, T_UNIQUE, T_FORALL}
CONS_TYPES = {CONS_INT, CONS_OPT}


def def_constraint(
    compiler: SatCompiler,
    constype: String,
    name: Var,
    loops: list,
    expr: Expr,
    level: Number,
):
    """
    Parameters
    ----------
    compiler : SatCompiler
            Current running compiler instance.
    constype : String
            Either 'opt' or 'int'.
    name : Var
            Constraint identifier.
    loops : list
            Nested replication loops.
    expr : Expr
            Inner constraint expression.
    level : Number, Var
            Constraint's penalty level.
    """

    # -*- Penalty Level -*-
    if level is None:
        pass
    elif isinstance(level, Number):
        pass
    elif isinstance(level, (Var, Expr)):
        level = compiler.evaluate(level, miss=True, calc=True, null=True)
    else:
        raise TypeError(
            f"'level' must be 'Number', 'Var' or 'Expr', not ({type(level)})"
        )

    # -*- Constraint Type & Level Consistency Check -*-
    if str(constype) not in CONS_TYPES:
        compiler << SatTypeError(
            f"Invalid constraint type '{constype}'. Options are: {CONS_TYPES}",
            target=constype,
        )

    if str(constype) == CONS_OPT:
        if level is None:
            pass
        elif level < 0 or not level.is_int:
            compiler << SatValueError(
                "Penalty level must be a positive integer", target=level
            )

    if str(constype) == CONS_INT:
        if level is None:
            compiler << SatValueError(
                "A Penalty level must be given to every Integrity Constraint",
                target=name,
            )
        elif level < 0 or not level.is_int:
            compiler << SatValueError(
                "Penalty level must be a positive integer", target=level
            )

    compiler.checkpoint()

    # -*- Quantifier Replication Stack -*-
    stack = Stack()

    # -*- Variable Accountance -*-
    var_bag = set()

    # -*- Stacking Quantifiers -*-
    l_type: String
    l_var: Var
    l_bounds: tuple[Number, Number, Number]
    l_conds: list[Callable]

    for (l_type, l_var, l_bounds, l_conds) in loops:
        if l_var in var_bag:
            compiler << SatExprError(
                f"Loop variable '{l_var}' repetition.", target=l_var
            )
        elif l_var in compiler:
            compiler < SatWarning(
                f"Variable '{l_var}' redefinition. Global values are set into the expression before loop variables.",
                target=l_var,
            )

        compiler.checkpoint()

        var_bag.add(l_var)

        # -*- Extract Loop Bounds -*-
        start, stop, step = l_bounds

        start = compiler.evaluate(start, miss=True, calc=True)
        stop = compiler.evaluate(stop, miss=True, calc=True)
        step = compiler.evaluate(step, miss=True, calc=True, null=True)

        if step is None:
            if start <= stop:
                step = Number("1")
            else:
                step = Number("-1")

        elif step == Number("0"):
            compiler << SatValueError("Step must be non-zero.", target=step)

        if ((start < stop) and (step < 0)) or ((start > stop) and (step > 0)):
            compiler << SatValueError("Inconsistent loop definition", target=start)

        compiler.checkpoint()

        # -*- Condition Compilation -*-
        if l_conds is not None:
            # Compile condition expressions
            cond = reduce(
                lambda x, y: x._AND_(y),
                [
                    compiler.evaluate(
                        c, miss=True, calc=True, context={var: var for var in var_bag}
                    )
                    for c in l_conds
                ],
                Number("1"),
            )
        else:
            cond = None

        # -*- Stack in reverse order -*-
        stack.push(
            {"type": l_type, "var": l_var, "bounds": (start, stop, step), "cond": cond}
        )

    # -*- Expression Analysis -*-
    if stdlog[3]:
        stdlog[3] << f"\n@{name}(raw):\n\t{expr}"

    if str(constype) == CONS_INT and expr.is_expr and not expr.logical:
        compiler << SatExprError(
            "Integrity constraint expressions must be purely logical i.e. no arithmetic operations allowed.",
            target=expr,
        )

    compiler.checkpoint()

    # Checks for undefined variables + evaluate constants
    # Adds inner scope where inside-loop variables point to themselves. This prevents both
    # older variables from interfering in evaluation and also undefined loop variables.
    # Last but not least, automatically removes this artificial scope
    expr: Expr = compiler.source.propagate(
        expr,
        compiler.evaluate(
            expr,
            miss=True,
            calc=False,
            null=False,
            context={var: var for var in var_bag},
        ),
        out=True,
    )

    if stdlog[3]:
        stdlog[3] << f"\n@{name}(verified):\n\t{expr}"

    # Simplifies expression
    expr: Expr = compiler.source.propagate(
        expr,
        compiler.evaluate(expr, miss=False, calc=True, null=False),
        out=True,
    )

    if stdlog[3]:
        stdlog[3] << f"\n@{name}(simplified):\n\t{expr}"

    build(compiler, constype, stack, Stack(), level, expr)


def build(
    compiler: SatCompiler, constype: str, A: Stack, B: Stack, level: Number, expr: Expr
):
    """"""
    if A:
        # -*- Retrieve Innermost Quatifier -*-
        item = A.pop()

        if constype == CONS_INT:
            if item["type"] == T_UNIQUE:
                # -*- Winner Takes All -*-
                R = A.copy()
                S = B.copy()

                B.push({**item, "type": T_EXISTS})
                build(compiler, constype, A, B, level, expr)

                var = Var(f"${item['var']}")

                if item["cond"] is None:
                    cond = Expr(T_NE, item["var"], var)
                else:
                    cond = Expr(
                        T_AND,
                        Expr.sub(item["cond"], item["var"], var),
                        Expr(T_NE, item["var"], var),
                    )

                S.push(
                    {
                        "type": T_EXISTS,
                        "var": var,
                        "bounds": item["bounds"],
                        "cond": cond,
                    }
                )
                S.push({**item, "type": T_EXISTS})

                build(
                    compiler,
                    constype,
                    R,
                    S,
                    level + 1,
                    Expr(
                        T_NOT, Expr(T_AND, expr, Expr.sub(expr, item["var"], var))
                    ).dnf,
                )
            elif item["type"] == T_FORALL:
                B.push({**item, "type": T_EXISTS})
                build(compiler, constype, A, B, level, expr)
            elif item["type"] == T_EXISTS:
                B.push({**item, "type": T_FORALL})
                build(compiler, constype, A, B, level, expr)
            else:
                raise ValueError(f"Invalid quatifier {item['type']}")
        elif constype == CONS_OPT:
            if item["type"] == T_UNIQUE:
                # -*- Winner Takes All -*-
                R = A.copy()
                S = B.copy()

                B.push({**item, "type": T_FORALL})
                build(compiler, constype, A, B, level, expr)

                var = Var(f"${item['var']}")

                if item["cond"] is None:
                    cond = Expr(T_NE, item["var"], var)
                else:
                    cond = Expr(
                        T_AND,
                        Expr.sub(item["cond"], item["var"], var),
                        Expr(T_NE, item["var"], var),
                    )

                S.push(
                    {
                        "type": T_FORALL,
                        "var": var,
                        "bounds": item["bounds"],
                        "cond": cond,
                    }
                )
                S.push({**item, "type": T_FORALL})

                build(
                    compiler,
                    constype,
                    R,
                    S,
                    level + 1,
                    Expr(
                        T_NOT, Expr(T_AND, expr, Expr.sub(expr, item["var"], var))
                    ).dnf,
                )
            elif item["type"] == T_FORALL:
                B.push(item)
                build(compiler, constype, A, B, level, expr)
            elif item["type"] == T_EXISTS:
                B.push(item)
                build(compiler, constype, A, B, level, expr)
            else:
                raise ValueError(f"Invalid quatifier {item['type']}")
        else:
            raise ValueError(f"Invalid constraint type '{constype}'")
    else:
        if constype == CONS_INT:
            expr = Expr.calculate(~expr)
            if expr.is_expr:
                energy = unstack(compiler, B, expr.dnf, {})
            else:
                # Here I'm supposing that everything that is not an 'Expr' is in the C.N.F.
                # (Also in any other normal form, such as D.N.F.)
                energy = unstack(compiler, B, expr, {})
        elif constype == CONS_OPT:
            if expr.is_expr:
                energy = unstack(compiler, B, expr.dnf, {})
            else:
                energy = unstack(compiler, B, expr, {})
        else:
            raise ValueError(f"Invalid constraint type '{constype}'")

        if level is None:
            level = 0

        if stdlog[3]:
            stdlog[3] << energy

        compiler.constraints[constype].append((int(level), energy))


def unstack(
    compiler: SatCompiler, stack: Stack, expr: Expr, context: dict
) -> tuple[int, Posiform]:
    """"""

    if stack:
        # -*- Retrieve Innermost Quatifier -*-
        item = stack.pop()

        var: Var = item["var"]

        if item["type"] == T_FORALL:
            energy = Posiform(1.0)
            for i in arange(*item["bounds"]):
                context[var] = i
                if item["cond"] is None or compiler.evaluate(
                    item["cond"], miss=False, context=context
                ):
                    energy *= unstack(compiler, stack, expr, context)

            del context[var]

        elif item["type"] == T_EXISTS:
            energy = Posiform(0.0)
            for i in arange(*item["bounds"]):
                context[var] = i
                if item["cond"] is None or compiler.evaluate(
                    item["cond"], miss=False, context=context
                ):
                    energy += unstack(compiler, stack, expr, context)

            del context[var]
        else:
            raise ValueError(f"Invalid Quantifier '{item['type']}'")

        # -*- Get Quantifier Back -*-
        stack.push(item)

        return energy
    else:
        return H(compiler, compiler.evaluate(expr, miss=False, context=context))


def H(compiler: SatCompiler, x: SatType) -> Posiform:
    """Energy Equation Mapping"""

    if x.is_expr:
        if x.head == T_NOT:
            return 1.0 - H(compiler, x[1])
        elif x.head == T_AND or x.head == T_MUL:
            e = Posiform(1.0)
            for y in x.tail:
                e *= H(compiler, y)
            return e
        elif x.head == T_POW:
            a, b = x.tail
            try:
                return H(compiler, a) ** H(compiler, b)
            except TypeError:
                compiler << SatValueError(
                    f"Can't compute power to non-integer '{b}'", target=b
                )
        elif x.head == T_OR or x.head == T_ADD:
            e = Posiform(0.0)
            for y in x.tail:
                e += H(compiler, y)
            return e
        elif x.head == T_XOR:
            a, b = x.tail
            return H(compiler, (a & ~b) | (~a & b))
        elif x.head == T_NEG:
            if len(x.tail) == 1:
                (a,) = x.tail
                return -H(compiler, a)
            elif len(x.tail) == 2:
                a, b = x.tail
                return H(compiler, a) - H(compiler, b)
            else:
                raise ValueError(
                    f"Unable to map '{x.head}({len(x.tail)})' into energy equation"
                )
        else:
            raise ValueError(f"Unable to map '{x.head}' into energy equation")
    elif x.is_number:
        return Posiform(x)
    elif x.is_var:
        return Posiform({(x,): 1.0})
    elif x.is_array:
        compiler << SatIndexError(f"Partially indexed array '{x}'", target=x)
    else:
        raise ValueError(f"Unable to map '{x}' into energy equation")

    compiler.checkpoint()



# Standard Library
import decimal
import numbers
import re

# Local
from .base import SatType
from ..error import SatValueError
from ..satlib import Source


class Number(decimal.Decimal, numbers.Real, SatType):
    """"""

    context = decimal.getcontext()
    context.prec = 32

    regex = re.compile(r"(\.[0-9]*[1-9])(0+)|(\.0*)$")

    def __new__(cls, value: numbers.Real, source: Source = None, lexpos: int = None):
        return decimal.Decimal.__new__(cls, value)

    def __init__(self, value: numbers.Real, source: Source = None, lexpos: int = None):
        SatType.__init__(self, source=source, lexpos=lexpos)

    def __hash__(self) -> int:
        return decimal.Decimal.__hash__(self)

    @classmethod
    def numeric(cls, x: object) -> bool:
        return isinstance(x, (cls, numbers.Real))

    def __str__(self):
        return self.regex.sub(r"\1", decimal.Decimal.__str__(self))

    def __repr__(self):
        return f"Number('{self.__str__()}')"

    def __abs__(self):
        return Number(decimal.Decimal.__abs__(self))

    def __int__(self):
        return decimal.Decimal.__int__(self)

    def __float__(self):
        return decimal.Decimal.__float__(self)

    ## Comparison
    def _GT_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__gt__(self, other))
        else:
            return NotImplemented

    def _GE_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__ge__(self, other))
        else:
            return NotImplemented

    def _LT_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__lt__(self, other))
        else:
            return NotImplemented

    def _LE_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__le__(self, other))
        else:
            return NotImplemented

    def _EQ_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__eq__(self, other))
        else:
            return NotImplemented

    def _NE_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__ne__(self, other))
        else:
            return NotImplemented

    ## Arithmetic
    def _ADD_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__add__(self, other))
        elif self == Number(0):
            return other
        else:
            return NotImplemented

    def _RADD_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__radd__(self, other))
        elif self == Number(0):
            return other
        else:
            return NotImplemented

    def _SUB_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__sub__(self, other))
        elif self == Number(0):
            return -other
        else:
            return NotImplemented

    def _RSUB_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__rsub__(self, other))
        elif self == Number(0):
            return other
        else:
            return NotImplemented

    def _MUL_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__mul__(self, other))
        elif self == Number(0):
            return Number(0)
        elif self == Number(1):
            return other
        else:
            return NotImplemented

    def _RMUL_(self, other):
        if self.numeric(other):
            return Number(decimal.Decimal.__rmul__(self, other))
        elif self == Number(0):
            return Number(0)
        elif self == Number(1):
            return other
        else:
            return NotImplemented

    def _DIV_(self, other):
        if self.numeric(other):
            if other != Number(0):
                return Number(decimal.Decimal.__truediv__(self, other))
            else:
                raise SatValueError("Division by zero.", target=other)
        elif self == Number(0):
            return Number(0)
        else:
            return NotImplemented

    def _RDIV_(self, other):
        if self == Number(0):
            raise SatValueError("Division by zero.", target=self)
        elif self == Number(1):
            return other
        elif self.numeric(other):
            return Number(decimal.Decimal.__truediv__(other, self))
        else:
            return NotImplemented

    def _MOD_(self, other):
        if self.numeric(other):
            if other != Number(0):
                return Number(decimal.Decimal.__mod__(self, other))
            else:
                raise SatValueError("Division by zero.", target=other)
        elif self == Number(0):
            return Number(0)
        else:
            return NotImplemented

    def _RMOD_(self, other):
        if self == Number("0"):
            raise SatValueError("Division by zero.", target=self)
        elif self.numeric(other):
            return Number(decimal.Decimal.__truediv__(other, self))
        else:
            return NotImplemented

    def _AND_(self, other):
        if self.numeric(other):
            return Number(self._MUL_(other))
        elif self == Number(1):
            return other
        elif self == Number(0):
            return self
        else:
            return NotImplemented

    def _RAND_(self, other):
        if self.numeric(other):
            return Number(other._MUL_(self))
        elif self == Number(1):
            return other
        elif self == Number(0):
            return self
        else:
            return NotImplemented

    def _OR_(self, other):
        if self.numeric(other):
            return Number((self._ADD_(other))._SUB_(self._MUL_(other)))
        elif self == Number(1):
            return self
        elif self == Number(0):
            return other
        else:
            return NotImplemented

    def _ROR_(self, other):
        if self.numeric(other):
            return Number((self._ADD_(other))._SUB_(self._MUL_(other)))
        elif self == Number(1):
            return self
        elif self == Number(0):
            return other
        else:
            return NotImplemented

    def _XOR_(self, other):
        if self.numeric(other):
            return Number(
                (self._ADD_(other))._SUB_(Number("2")._MUL_(self)._MUL_(other))
            )
        else:
            return NotImplemented

    def _RXOR_(self, other):
        if self.numeric(other):
            return Number(
                (self._ADD_(other))._SUB_(Number("2")._MUL_(self)._MUL_(other))
            )
        else:
            return NotImplemented

    def _IMP_(self, other):
        if self.numeric(other):
            return Number((self._NOT_())._OR_(other))
        elif self == Number(0):
            return Number(1)
        elif self == Number(1):
            return other
        else:
            return NotImplemented

    def _RIMP_(self, other):
        if self.numeric(other):
            return Number(self._OR_(other._NOT_()))
        elif self == Number(1):
            return self
        elif self == Number(0):
            return other._NOT_()
        else:
            return NotImplemented

    def _IFF_(self, other):
        if self.numeric(other):
            return Number((self._NOT_())._AND_(other._NOT_())._OR_(self._AND_(other)))
        elif self == Number(1):
            return other
        elif self == Number(0):
            return other._NOT_()
        else:
            return NotImplemented

    def _NOT_(self):
        return Number(Number(1)._SUB_(self))

    def _NEG_(self):
        return Number(decimal.Decimal.__neg__(self))

    ## Aliases
    def __add__(self, other):
        return self._ADD_(other)

    def __sub__(self, other):
        return self._SUB_(other)

    def __neg__(self):
        return self._NEG_()

    def __mul__(self, other):
        return self._MUL_(other)

    def __eq__(self, other):
        return self._EQ_(other)

    def __ne__(self, other):
        return self._NE_(other)

    def __lt__(self, other):
        return self._LT_(other)

    def __le__(self, other):
        return self._LE_(other)

    def __gt__(self, other):
        return self._GT_(other)

    def __ge__(self, other):
        return self._GE_(other)

    @classmethod
    def prec(cls, value: int = None) -> int:
        if value is not None:
            cls.context.prec = value
        return int(cls.context.prec)

    # -*- Type Checking -*-
    @property
    def is_int(self) -> bool:
        return self == decimal.Decimal.__floordiv__(self, 1)

    @property
    def is_array(self) -> bool:
        return False

    @property
    def is_var(self) -> bool:
        return False

    @property
    def is_number(self) -> bool:
        return True

    @property
    def is_expr(self) -> bool:
        return False

__all__ = ["Number"]



from .sys_config import sys_config
from .def_array import def_array
from .def_constant import def_constant
from .def_constraint import def_constraint
from .run_init import run_init
from .run_script import run_script
from ...symbols import SYS_CONFIG, DEF_ARRAY, DEF_CONSTANT, DEF_CONSTRAINT, RUN_SCRIPT, RUN_INIT

INSTRUCTIONS = {
    RUN_INIT: run_init,
    SYS_CONFIG: sys_config,
    DEF_ARRAY: def_array,
    DEF_CONSTANT: def_constant,
    DEF_CONSTRAINT: def_constraint,
    RUN_SCRIPT: run_script
}

__all__ = ["INSTRUCTIONS"]


""" DEF_ARRAY
    =========

    STATUS: OK
"""

from ..compiler import SatCompiler
from ...types import Var, Array, Number, PythonObject
from ...error import SatTypeError, SatValueError

def def_array(compiler: SatCompiler, name: Var, shape: tuple, buffer: list):
    """ Array definition
        ================

        1. Evaluate every shape component
        2. Evaluate buffer content
        3. Set Array to memory
    """
    ## Evaluates shape values and variables
    shape = tuple(compiler.evaluate(n, miss=True, calc=True, track=True) for n in shape)

    compiler.checkpoint()
    
    ## Checks for array shape consistency.
    def_array_shape(compiler, shape)

    ## Special treatment for python object buffer
    if type(buffer) is PythonObject:
        def_python_buffer(compiler, name, shape, buffer)
    else:
        ## Runs correction checking and normalization
        def_array_buffer(compiler, name, shape, buffer)

    compiler.checkpoint()

def def_array_shape(compiler: SatCompiler, shape: tuple):   
    """ DEF_ARRAY_SHAPE
        ===============
    """
    for n in shape:
        if type(n) is not Number or not (n.is_int and n > 0):
            compiler << SatTypeError(f'Array dimensions must be positive integers.', target=n)

def def_python_buffer(compiler: SatCompiler, name: Var, shape: tuple, buffer: PythonObject):
    """ DEF_ARRAY_BUFFER
        ================
    """
    ## retrieve object
    if type(buffer.obj) is list:
        compiler.memset(name, Array.from_list(name, shape, buffer))
    elif type(buffer.obj) is dict:
        compiler.memset(name, Array.from_dict(name, shape, buffer))
    else:
        compiler << SatTypeError()

def def_array_buffer(compiler: SatCompiler, name: Var, shape: tuple, buffer: list):
    """ DEF_ARRAY_BUFFER
        ================
    """

    array = {}

    if buffer is None:
        pass
    else:
        for idx, val in buffer:
            idx = tuple(compiler.evaluate(i, miss=True, calc=True) for i in idx)

            if len(idx) > len(shape):
                compiler << SatValueError(f'Too much indices for {len(shape)}-dimensional array', target=idx[len(shape)])
                
            for i, n in zip(idx, shape):
                if not i.is_int:
                    compiler << SatTypeError(f'Array indices must be integers.', target=i)
                    
                if not (1 <= i <= n):
                    compiler << SatValueError(f'Indexing ´{i}´ is out of bounds [1, {n}]', target=i)

            val = compiler.evaluate(val, miss=True, calc=True)

            if type(val) is not Number:
                compiler << SatValueError(f'Array elements must be numbers.', target=val)
            else:
                array[tuple(map(int, idx))] = val

        compiler.checkpoint()

    ## Sets new `types.Array` to memory.
    ## Also runs automatic compiler.checkpoint()
    compiler.memset(name, Array.from_buffer(name, shape, array))


""""""
# Future
from __future__ import annotations

# Standard Library
import json
import numbers

# Third-Party
import numpy as np


class Posiform(dict):
    r"""
    This object is intended to represent a sum of products under a psudo-boolean domain.

    Assumptions
    -----------
    1. Every variable is boolean (i.e. $x \in \{0, 1\}$)
    2. Posiform is a mutable type.
    """

    def __init__(self, buffer: dict | float = None):
        """\
        Parameters
        ----------
        buffer : dict
            Dictionary containing pairs (variables, constant).
        """
        if buffer is not None:
            if isinstance(buffer, dict):
                __buffer = {}
                for k, v in buffer.items():
                    if not isinstance(v, numbers.Real):
                        raise TypeError("Posiform buffer values must be real numbers")
                    elif float(v) == 0.0:
                        continue

                    if k is None:
                        __buffer[k] = float(v)
                    else:
                        if not isinstance(k, (tuple, set, frozenset)) or not all(isinstance(s, str) for s in k):
                            raise TypeError("Posiform buffer keys must be 'None' or be of type 'tuple' or 'set' and contain only 'strings'.")
                        else:
                            __buffer[frozenset(map(str, k))] = float(v)
            elif isinstance(buffer, numbers.Real):
                if buffer == 0.0:
                    __buffer = {}
                else:
                    __buffer = {None: float(buffer)}
            else:
                raise TypeError(f"Posiform buffer must be of type 'dict', 'float', 'int' or 'None', not '{type(buffer)}'")
        else:
            __buffer = {}
        dict.__init__(self, __buffer)
        self.__aux = 0

    def __bool__(self) -> bool:
        return len(self) > 0

    def __str__(self) -> str:
        if self:
            terms = []
            for k, v in self:
                if k is None:
                    if v >= 0.0:
                        terms.extend(["+", str(v)])
                    else:
                        terms.extend(["-", str(abs(v))])
                else:
                    if v == 1.0:
                        terms.extend(["+", " * ".join(sorted(k))])
                    elif v == -1.0:
                        terms.extend(["-", " * ".join(sorted(k))])
                    elif v >= 0.0:
                        terms.extend(["+", " * ".join([str(v), *sorted(k)])])
                    else:
                        terms.extend(["-", " * ".join([str(abs(v)), *sorted(k)])])
            return " ".join(terms[1:] if terms[0] == "+" else terms)
        else:
            return "0.0"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict.__repr__({k if k is None else tuple(sorted(k)) : v for k, v in self})})"

    def __iter__(self):
        return iter(dict.items(self))

    def copy(self) -> Posiform:
        """Deep Posiform copy."""
        return Posiform({k: v for k, v in self})

    def __float__(self) -> float:
        if len(self) == 0:
            return 0.0
        elif len(self) == 1 and None in self:
            return self[None]
        else:
            raise TypeError("Can't cast 'Posiform' to 'float' due to non-constant terms")
            

    def __call__(self, point: dict) -> Posiform:
        """"""

        if isinstance(point, dict):

            posiform = Posiform()

            term: frozenset | None
            cons: float

            for term, cons in self:
                if term is None:
                    if None in posiform:
                        posiform[None] += cons
                    else:
                        posiform[term] = cons
                    continue
                
                for x in point:
                    if not isinstance(x, str):
                        raise TypeError(f"Variables must be of type 'str', not '{type(x)}'")

                    c = point[x]

                    if isinstance(c, str):
                        if x in term:
                            term -= {x}
                            term |= {c}
                    elif isinstance(c, numbers.Real):
                        if x in term:
                            term -= {x}
                            cons *= c
                    else:
                        raise TypeError(f"Evaluation point coordinates must be either real numbers ('int', 'float') or variables ('str'), not '{type(c)}'")

                if not term:
                    if None in posiform:
                        posiform[None] += cons
                    else:
                        posiform[None] = cons
                elif term in posiform:
                    posiform[term] += cons
                else:
                    posiform[term] = cons

            return posiform
        else:
            raise TypeError(f"Can't evaluate Posiform at non-mapping {point} of type {type(point)}")

    def __add__(self, other) -> Posiform:
        posiform = self.copy()
        posiform = posiform.__iadd__(other)
        return posiform

    def __iadd__(self, other) -> Posiform:
        if isinstance(other, dict):
            try:
                other = self.__class__(other)
            except TypeError as type_error:
                raise TypeError("Unable to cast operand to Posiform type.") from type_error
        if isinstance(other, type(self)):
            for k, v in other:
                if k in self:
                    w = self[k] + v
                    if w == 0.0:
                        del self[k]
                    else:
                        self[k] = w
                else:
                    self[k] = v
            return self
        elif isinstance(other, numbers.Number) and not isinstance(other, complex):
            if None in self:
                self[None] += float(other)
            else:
                self[None] = float(other)
            return self
        else:
            return NotImplemented

    def __neg__(self) -> Posiform:
        posiform = self.__class__()
        for k, v in self:
            posiform[k] = -v
        return posiform

    def __rsub__(self, other) -> Posiform:
        return self.__sub__(other).__neg__()

    def __sub__(self, other) -> Posiform:
        posiform = self.copy()
        posiform = posiform.__isub__(other)
        return posiform

    def __isub__(self, other) -> Posiform:
        if isinstance(other, dict):
            try:
                other = self.__class__(other)
            except TypeError as type_error:
                raise TypeError("Unable to cast operand to Posiform type.") from type_error
        if isinstance(other, type(self)):
            for k, v in other.items():
                if k in self:
                    w = self[k] - v
                    if w == 0.0:
                        del self[k]
                    else:
                        self[k] = w
                else:
                    self[k] = -v
            return self
        elif isinstance(other, numbers.Number) and not isinstance(other, complex):
            if None in self:
                self[None] -= float(other)
            else:
                self[None] = -float(other)
            return self
        else:
            return NotImplemented

    def __truediv__(self, other) -> Posiform:
        posiform = self.copy()
        posiform = posiform.__itruediv__(other)
        return posiform

    def __itruediv__(self, other) -> Posiform:
        if isinstance(other, numbers.Number):
            c = float(other)
            if c != 0.0:
                for k, v in self:
                    self[k] = v / c
                return self
            else:
                raise ZeroDivisionError("division by zero")
        else:
            return NotImplemented

    def __pow__(self, other) -> Posiform:
        posiform = self.copy()
        posiform = posiform.__ipow__(other)
        return posiform

    def __ipow__(self, other) -> Posiform:
        if isinstance(other, dict):
            try:
                other = self.__class__(other)
            except TypeError as type_error:
                raise TypeError("Unable to cast operand to 'Posiform' type") from type_error
        if isinstance(other, type(self)) or isinstance(other, numbers.Real):
            try:
                other = float(other)
            except TypeError as type_error:
                raise TypeError("Unable to cast operand to 'float' type") from type_error

            if not other.is_integer() or other < 0:
                raise TypeError("Can't raise 'Posiform' to non-integer power")
            else:
                other = int(other)

            posiform = Posiform(1.0)
            while other > 0:
                posiform *= self
                other -= 1
            return posiform
        else:
            return NotImplemented

    def __mul__(self, other) -> Posiform:
        posiform = self.copy()
        posiform = posiform.__imul__(other)
        return posiform

    def __imul__(self, other):
        if isinstance(other, dict):
            try:
                other = self.__class__(other)
            except TypeError as type_error:
                raise TypeError("Unable to cast operand to Posiform type") from type_error
        if isinstance(other, type(self)):
            posiform = Posiform()
            for kx, vx in self:
                for ky, vy in other:
                    if kx is None:
                        k = ky
                    elif ky is None:
                        k = kx
                    else:
                        k = frozenset({*kx, *ky})

                    v = vx * vy

                    if k in posiform:
                        w = posiform[k] + v
                        if w == 0.0:
                            del posiform[k]
                        else:
                            posiform[k] = w
                    else:
                        posiform[k] = v
            return posiform
        elif isinstance(other, numbers.Real):
            w = float(other)
            if w == 0.0:
                self.clear()
            else:
                for k, v in self:
                    self[k] = w * v
            return self
        else:
            return NotImplemented

    __radd__ = __add__
    __rmul__ = __mul__

    @property
    def aux(self) -> str:
        self.__aux += 1
        return f"${self.__aux}"

    def reduce_degree(self) -> Posiform:
        """"""
        ## Reset ancillary variable counter
        self.__aux = 0

        posiform = Posiform()
        for X, a in self:
            posiform += self.__reduce_term(X, a)
        return posiform

    __RED_NO = 0 # No Reduction
    __RED_MS = 1 # Minimum Selection
    __RED_SU = 2 # Substitution

    def __reduce_strategy(self, term: frozenset, __cons: float) -> int:
        """"""
        if term is None or len(term) <= 2:
            return self.__RED_NO
        else:
            return self.__RED_MS

    def __reduce_term(self, term: frozenset, cons: float) -> Posiform:
        """
        Parameters
        ----------
        term : frozenset[str], tuple[str], list[str], set[str]
            Tuple with the string variable names (or None for constant term).
        cons : float
            Multiplicative constant.

        Returns
        -------
        Posiform
        """
        strategy = self.__reduce_strategy(term, cons) 

        if strategy == self.__RED_NO:
            return self.__class__({term: cons})
        elif strategy == self.__RED_MS:
            return self.__minimum_selection(term, cons)
        elif strategy == self.__RED_SU:
            return self.__substitution(term, cons)
        else:
            raise NotImplementedError("No more strategies were implemented")

    def __substitution(self, X: frozenset, a: float) -> Posiform:
        """"""
        w = self.aux
        x, y, *z = X
        # TODO: How can I compute alpha? (besides alpha > 1)
        alpha = 2.0
        return a * (self.__reduce_term((*z, w), 1.0) + alpha * self.__P(x, y, w))

    def __minimum_selection(self, X: frozenset, a: float) -> Posiform:
        """"""
        w = self.aux
        x, y, *z = X
        if a < 0:
            ## if a < 0: a (x y z) => a w (x + y + z - 2)
            return self.__class__({(x, w): a, (y, w): a, (w,): -2.0 * a}) + self.__reduce_term((*z, w), a)
        else:
            ## if a > 0: a (x y z) => a (x w + y w + z w + x y + x z + y z - x - y - z - w + 1)
            return (
                self.__class__(
                    {
                        (x, w): a,
                        (y, w): a,
                        (x, y): a,
                        (x,): -a,
                        (y,): -a,
                        (w,): -a,
                        None: a,
                    }
                )
                + (self.__reduce_term((x, *z), a) + self.__reduce_term((y, *z), a) + self.__reduce_term((*z, w), a) + self.__reduce_term((*z,), -a))
            )

    @classmethod
    def __P(cls: type, x: str, y: str, w: str):
        """"""
        return cls({(x, y): 1.0, (x, w): -2.0, (y, w): -2.0, (w,): 3.0})

    @property
    def variables(self) -> list:
        return sorted(set(x for k in self.keys() if k is not None for x in k))

    @property
    def degree(self) -> int:
        return max(map(lambda k: 0 if k is None else len(k), self))

    def toMiniJSON(self) -> str:
        return json.dumps({("" if k is None else " ".join(sorted(k))): v for k, v in self})

    @classmethod
    def fromMiniJSON(cls, data: str) -> Posiform:
        """"""
        json_data = json.loads(data)

        buffer = {}

        if not isinstance(json_data, dict):
            raise json.decoder.JSONDecodeError("Data must be of Object type (Python dict)", data, 0)
        
        for term, cons in json_data.items():
            if not isinstance(cons, (int, float)):
                raise json.decoder.JSONDecodeError("Constants must be of Number type (Python int, float)", data, 0)
            else:
                if term != "":
                    key = frozenset(term.split(" "))
                else:
                    key = None
                val = float(cons)

                if key in buffer:
                    buffer[key] += val
                else:
                    buffer[key] = val
        return cls(buffer)

    def toJSON(self, indent: int = None) -> str:
        if indent is None:
            return json.dumps([{"term": sorted(k) if k is not None else None, "cons": v} for k, v in self])
        else:
            return json.dumps([{"term": sorted(k) if k is not None else None, "cons": v} for k, v in self], indent=indent)

    @classmethod
    def fromJSON(cls, data: str) -> Posiform:
        """"""
        json_data = json.loads(data)

        buffer = {}

        if not isinstance(json_data, list):
            raise json.decoder.JSONDecodeError("Data must be of Array type (Python list)", data, 0)
        
        for item in json_data:
            if not isinstance(item, dict):
                raise json.decoder.JSONDecodeError("Items must be of Object type (Python dict)", data, 0)
            elif "term" not in item or "cons" not in item:
                raise json.decoder.JSONDecodeError("Items must contain both 'term' and 'cons' keys", data, 0)
            elif not isinstance(item["term"], list) and item["term"] is not None:
                raise json.decoder.JSONDecodeError("Items 'term' value must be of Array type (Python list)", data, 0)
            elif item["term"] is not None and not all(isinstance(var, str) for var in item["term"]):
                raise json.decoder.JSONDecodeError("Items 'term' value must contain only entries of type String (Python str)", data, 0)
            elif not isinstance(item["cons"], (int, float)):
                raise json.decoder.JSONDecodeError("Items 'cons' value must be of Number type (Python float)", data, 0)
            else:
                key = tuple(item["term"]) if item["term"] is not None else None
                val = float(item["cons"])
                if key in buffer:
                    buffer[key] += val
                else:
                    buffer[key] = val
        return cls(buffer)

    def qubo(self) -> tuple[dict[str, int], np.ndarray, float]:
        """
        Returns
        -------
        dict[str, int]
            Mapping between variables and respective indexes.
        np.ndarray
            Symmetric Matrix representing QUBO instance.
        float
            Ground state energy.
        """
        reduced = self.reduce_degree()
        variables = reduced.variables

        n = len(variables)

        x = {v: i for i, v in enumerate(variables)}
        Q = np.zeros((n, n), dtype=float)
        c = 0.0

        for k, a in reduced:
            if k is None:
                c = a
            else:   
                I = tuple(map(x.get, k))
                if len(I) == 1:
                    i = I[0]
                    Q[i, i] += a
                elif len(I) == 2:
                    i, j = I
                    Q[i, j] += a / 2.0
                    Q[j, i] += a / 2.0
                else:
                    raise RuntimeError("Degree reduction failed.")
        return x, Q, c

    def ising(self) -> tuple[dict[str, int], np.ndarray, float]:
        """
        Returns
        -------
        dict[str, int]
            Mapping between variables and respective indexes.
        np.ndarray
            Symmetric Matrix representing QUBO instance.
        float
            Ground state energy.
        """
        reduced = self.reduce_degree()
        variables = reduced.variables

        n = len(variables)

        x = {v: i for i, v in enumerate(variables)}
        J = np.zeros((n, n), dtype=float)
        c = 0.0

        for k, a in reduced:
            if k is None:
                c = a
            else:   
                I = tuple(map(x.get, k))
                if len(I) <= 2:
                    J[I] += a
                else:
                    raise RuntimeError("Degree reduction failed.")
        return x, J, c


__all__ = ["Posiform"]



"""\
Satyrus Compiler v{version}
"""
from __future__ import annotations

__version__ = "3.0.7"

## Standard Library
import argparse
import json
from pathlib import Path
from functools import wraps
from gettext import gettext

## Third-Party
from cstream import Stream, stdlog, stdwar, stderr, stdout

## Local
from .help import satyrus_help
from ..api import SatAPI
from ..error import EXIT_SUCCESS, EXIT_FAILURE
from ..assets import SAT_BANNER, SAT_CRITICAL
from ..satlib import Timing, log
from ..satyrus import Satyrus


def load_json(parser, path: Path) -> object:
    if not path.exists() or not path.is_file():
        parser.error(f"File '{path}' doesn't exists")

    with path.open(mode="r") as file:
        try:
            data = json.load(file)
        except json.JSONDecodeError as error:
            parser.error(f"In '{path}':\n{error.msg}")
        else:
            return data


class ArgParser(argparse.ArgumentParser):
    @wraps(argparse.ArgumentParser.__init__)
    def __init__(self, *args, **kwargs):
        self._rank = {}
        argparse.ArgumentParser.__init__(self, *args, **kwargs)

    @wraps(argparse.ArgumentParser.parse_args)
    def parse_args(self, args: list = None, namespace: argparse.Namespace = None):
        args = argparse.ArgumentParser.parse_args(self, args, namespace)

        for p in sorted(self._rank.keys()):
            for action, params in self._rank[p]:
                action(*params)

        return args

    def print_help(self, from_help: bool = True):
        if from_help:
            stdlog[0] << SAT_BANNER
        argparse.ArgumentParser.print_help(self, stdlog[0])

    def error(self, message: str, code: int = 1):
        stderr << f"Error: {gettext(message)}"
        self.print_help(from_help=False)
        exit(code)

    @classmethod
    def enqueue(cls, priority: int = 0):
        def decor(action):
            @wraps(action)
            def enqueued_action(self, parser, namespace, values, option_string=None):
                if isinstance(parser, cls):
                    params = (self, parser, namespace, values, option_string)
                    if priority in parser._rank:
                        parser._rank[priority].append((action, params))
                    else:
                        parser._rank[priority] = [(action, params)]
                else:
                    return action(self, parser, namespace, values, option_string)

            return enqueued_action

        return decor


class GetVersion(argparse._VersionAction):
    @wraps(argparse._VersionAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._VersionAction.__init__(self, *args, **kwargs)

    @wraps(argparse._VersionAction.__call__)
    def __call__(self, parser, namespace, values, option_string=None):
        from sys import version_info

        (
            stdout[0]
            << f"satyrus {__version__} (python {version_info.major}.{version_info.minor})"
        )
        exit(EXIT_SUCCESS)


class SetVerbosity(argparse._StoreAction):
    @wraps(argparse._StoreAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(0)
    @wraps(argparse._StoreAction.__call__)
    def __call__(self, parser, namespace, values, option_string=None):
        argparse._StoreAction.__call__(self, parser, namespace, values, option_string)
        Stream.set_lvl(getattr(namespace, self.dest))


class DebugMode(argparse._StoreTrueAction):
    @wraps(argparse._StoreTrueAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreTrueAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(1)
    @wraps(argparse._StoreTrueAction.__call__)
    def __call__(self, parser, namespace, values, option_string=None):
        argparse._StoreTrueAction.__call__(
            self, parser, namespace, values, option_string
        )
        setattr(namespace, "verbose", 3)
        Stream.set_lvl(3)


class SetSolver(argparse._StoreAction):
    @wraps(argparse._StoreAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(3)
    @wraps(argparse._StoreAction.__call__)
    def __call__(self, parser: ArgParser, namespace, values, option_string=None):
        argparse._StoreAction.__call__(self, parser, namespace, values, option_string)
        solver = str.strip(getattr(namespace, self.dest))
        if solver not in SatAPI.options():
            parser.error(
                f"Invalid solver choice '{solver}' (choose from {SatAPI.options()!r})"
            )


class readParams(argparse._StoreAction):
    @wraps(argparse._StoreAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(4)
    @wraps(argparse._StoreAction.__call__)
    def __call__(self, parser: ArgParser, namespace, values, option_string=None):
        argparse._StoreAction.__call__(self, parser, namespace, values, option_string)
        params = getattr(namespace, self.dest)

        if params is None:
            stdwar[1] << "Warning: No solver parameters passed"
            setattr(namespace, self.dest, {})

        params_path = Path(params)

        params: dict = load_json(parser, params_path)

        if isinstance(params, dict):
            setattr(namespace, self.dest, params)
        else:
            parser.error(
                f"In '{params_path}':\nParameter object must be mapping, not '{type(params)}'"
            )

class readGuess(argparse._StoreAction):
    @wraps(argparse._StoreAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(4)
    @wraps(argparse._StoreAction.__call__)
    def __call__(self, parser: ArgParser, namespace, values, option_string=None):
        argparse._StoreAction.__call__(self, parser, namespace, values, option_string)
        
        guess_path = Path(getattr(namespace, self.dest))

        guess: dict = load_json(parser, guess_path)

        if guess is None:
            stdwar[1] << "Warning: No solver guess given"
        elif not isinstance(guess, dict):
            parser.error(
                f"In '{guess_path}':\nGuess object must be mapping, not '{type(guess)}'"
            )
        elif not all(v in {0, 1} for v in guess.values()):
            parser.error(f"In '{guess_path}':\nGuess values must be of type Number (Python int) and binary (0 or 1) ")
        elif not all(isinstance(k, str) for k in guess):
            parser.error(f"In '{guess_path}':\nGuess keys must be variables of type String (Python str)")

        setattr(namespace, self.dest, guess)

class SatCLI:
    r"""
      Satyrus Stack
      ┌───────────┐
    │ │  SatCLI   │
    │ ├───────────┤
    │ │  SatAPI   │
    │ ├───────────┤
    │ │  Satyrus  │
    │ ├───────────┤
    │ │SatCompiler│
    │ ├───────────┤
    ▼ │ SatParser │
      └───────────┘
    """

    @classmethod
    def infer_output(
        cls,
        ext: str,
        args: argparse.Namespace,
        answer: tuple[dict, float] | str | object,
    ) -> tuple[str, str]:
        if answer is None:
            exit(EXIT_FAILURE)

        if SatAPI.complete(answer):
            answer = json.dumps({"solution": answer[0], "energy": answer[1]}, indent=4)
            output = f"{Path(args.source[0]).stem}.out.json"
        else:
            answer = str(answer)
            output = f"{Path(args.source[0]).stem}.{ext}"

        if args.output is not None:
            output = args.output

        return output, answer

    @classmethod
    def run(cls, argv: list = None):
        """ """

        # -*- Set base output verbosity level -*-
        Stream.set_lvl(1)

        # -*- Load Interfaces -*-
        SatAPI._load()

        parser = ArgParser(
            prog="satyrus", description=__doc__.format(version=__version__)
        )

        ## Mandatory - Source file path
        parser.add_argument("source", help=satyrus_help("source"), nargs="+")

        # Optional - Compiler Version
        parser.add_argument(
            "--version", help=satyrus_help("version"), action=GetVersion
        )

        ## Optional - Compiler Verbosity
        parser.add_argument(
            "-v",
            "--verbose",
            type=int,
            dest="verbose",
            choices=[0, 1, 2, 3],
            default=1,
            help=satyrus_help("verbose"),
            action=SetVerbosity,
        )

        # Optional - Debug Mode
        parser.add_argument(
            "-d", "--debug", dest="debug", help=argparse.SUPPRESS, action=DebugMode
        )

        # Optional - Solver Option
        parser.add_argument(
            "-s",
            "--solver",
            type=str,
            dest="solver",
            help=satyrus_help("solver"),
            default="default",
            action=SetSolver,
        )

        # Optional - Solver Option
        parser.add_argument(
            "-c",
            "--clear",
            dest="clear",
            help=satyrus_help("clear"),
            action="store_true",
        )

        # Optional - Output File
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            dest="output",
            help=satyrus_help("output"),
            action="store",
        )

        # Optional - Legacy Syntax
        parser.add_argument(
            "-l",
            "--legacy",
            dest="legacy",
            action="store_true",
            help=satyrus_help("legacy"),
        )

        # Optional - Timing Report
        parser.add_argument(
            "-r",
            "--report",
            dest="report",
            action="store_true",
            help=satyrus_help("report"),
        )

        # Optional - Solver parameters
        parser.add_argument(
            "-p",
            "--params",
            dest="params",
            help=satyrus_help("params"),
            action=readParams,
        )

        # Optional - Augmented API
        parser.add_argument(
            "-g", "--guess", dest="guess", help=satyrus_help("guess"), action=readGuess
        )

        if argv is None:
            args = parser.parse_args()
        else:
            args = parser.parse_args(argv)

        # Exhibits Compiler Command line arguments
        stdlog[
            3
        ] << f'Command line args:\n{";".join(f"{k}={v!r}" for k, v in vars(args).items())}'

        # Warning: Debug mode
        if args.debug:
            stdwar[1] << "Warning: Debug mode enabled."

        # Warning: Legacy mode
        if args.legacy:
            stdwar[1] << "Warning: Parser is running in legacy mode."

        # Waning: Clear Cache
        if args.clear:
            Satyrus.clear_cache()
            stdwar[1] << "Warning: Cache Cleared."

        try:
            # Launch API
            api = SatAPI(*args.source, guess=args.guess, legacy=args.legacy)

            # Solve in desired way
            if args.params is None:
                params = {"$solver": args.solver}
            else:
                params = {"$solver": args.solver, **args.params}

            answer: tuple[dict, float] | str | object = api(**params)

            ext: str = api.extension(args.solver)

            # Retrieve Output Destination
            args.output, answer = cls.infer_output(ext, args, answer)

            with open(args.output, mode="w", encoding="utf8") as file:
                file.write(answer)

            # Report
            if args.report:
                Timing.timer.show_report(level=0)

            exit(EXIT_SUCCESS)
        except MemoryError as exc:
            stderr[0] << exc
            exit(EXIT_FAILURE)
        except RuntimeError as exc:
            stderr[0] << exc
            exit(EXIT_FAILURE)
        except Exception:
            trace = log(target="satyrus.log")
            if args.debug:
                stderr[0] << trace
            else:
                stderr[0] << SAT_CRITICAL
            exit(EXIT_FAILURE)


__all__ = ["SatCLI"]



"""\
Satyrus API v{version}
"""
from __future__ import annotations

__version__ = "3.0.7"

## Standard Library
import argparse
from functools import wraps
from gettext import gettext

## Third-Party
from cstream import Stream, stdlog, stderr, stdout

## Local
from .help import sat_api_help
from ..api import SatAPI
from ..error import EXIT_SUCCESS
from ..assets import SAT_BANNER

class ArgParser(argparse.ArgumentParser):
    @wraps(argparse.ArgumentParser.__init__)
    def __init__(self, *args, **kwargs):
        self._rank = {}
        argparse.ArgumentParser.__init__(self, *args, **kwargs)

    @wraps(argparse.ArgumentParser.parse_args)
    def parse_args(self, args: list = None, namespace: argparse.Namespace = None):
        args = argparse.ArgumentParser.parse_args(self, args, namespace)

        for p in sorted(self._rank.keys()):
            for action, params in self._rank[p]:
                action(*params)

        return args

    def print_help(self, *, from_help: bool = True):
        if from_help:
            stdlog[0] << SAT_BANNER
        argparse.ArgumentParser.print_help(self, stdlog[0])

    def error(self, message: str, code: int = 1):
        stderr << f"Error: {gettext(message)}"
        self.print_help(from_help=False)
        exit(code)

    @classmethod
    def enqueue(cls, priority: int = 0):
        def decor(action):
            @wraps(action)
            def enqueued_action(self, parser, namespace, values, option_string=None):
                if isinstance(parser, cls):
                    params = (self, parser, namespace, values, option_string)
                    if priority in parser._rank:
                        parser._rank[priority].append((action, params))
                    else:
                        parser._rank[priority] = [(action, params)]
                else:
                    return action(self, parser, namespace, values, option_string)

            return enqueued_action

        return decor

class GetVersion(argparse._VersionAction):
    @wraps(argparse._VersionAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._VersionAction.__init__(self, *args, **kwargs)

    @wraps(argparse._VersionAction.__call__)
    def __call__(self, parser, namespace, values, option_string=None):
        from sys import version_info

        stdout[0] << f"satyrus {__version__} (python {version_info.major}.{version_info.minor})"
        exit(EXIT_SUCCESS)

class SetVerbosity(argparse._StoreAction):
    @wraps(argparse._StoreAction.__init__)
    def __init__(self, *args, **kwargs):
        argparse._StoreAction.__init__(self, *args, **kwargs)

    @ArgParser.enqueue(0)
    @wraps(argparse._StoreAction.__call__)
    def __call__(self, parser, namespace, values, option_string=None):
        argparse._StoreAction.__call__(self, parser, namespace, values, option_string)
        Stream.set_lvl(getattr(namespace, self.dest))

def addSatAPI(args: argparse.Namespace):
    SatAPI.add(*args.path)

def buildSatAPI(args: argparse.Namespace):
    SatAPI.build(*args.path)

def removeSatAPI(args: argparse.Namespace):
    SatAPI.remove(*args.name, yes=args.yes)

def clearSatAPI(args: argparse.Namespace):
    SatAPI.clear(yes=args.yes)
        
class SatAPICLI:
    r"""
      Satyrus Stack
      ┌───────────┐
    │ │  SatCLI   │
    │ ├───────────┤
    │ │  SatAPI   │
    │ ├───────────┤
    │ │  Satyrus  │
    │ ├───────────┤
    │ │SatCompiler│
    │ ├───────────┤
    ▼ │ SatParser │
      └───────────┘
    """

    @classmethod
    def run(cls, argv: list = None):
        """
        Parameters
        ----------

        source : str
            Path to source code.
        -o, --out : str
            Output Destination.
        --legacy : bool
            If True, uses the legacy parser.
        -v, --verbose : int
            Compiler verbosity level.
        -d, --debug : bool
            If True, enables debug mode.
        -a, --api : str
            If present, include solver interfaces defined in python file.
        -p, --params : str
            Path to JSON file containing parameters for passing to Solver API.
        """

        parser = ArgParser(prog="sat-api", description=__doc__.format(version=__version__))
        parser.set_defaults(func=None)

        # Optional - Compiler Version
        parser.add_argument("--version", help=sat_api_help("version"), action=GetVersion)

        # Optional - Compiler Verbosity
        parser.add_argument(
            "-v",
            "--verbose",
            type=int,
            dest="verbose",
            choices=[0, 1, 2, 3],
            default=1,
            help=sat_api_help("verbose"),
            action=SetVerbosity,
        )

        # -*- Subparsers -*-
        subparsers = parser.add_subparsers()

        add_parser = subparsers.add_parser("add")
        add_parser.add_argument("path", help=sat_api_help("add"), nargs="+")
        add_parser.set_defaults(func=addSatAPI)

        build_parser = subparsers.add_parser("build")
        build_parser.add_argument("path", help=sat_api_help("build"), nargs="+")
        build_parser.set_defaults(func=buildSatAPI)

        remove_parser = subparsers.add_parser("remove")
        remove_parser.add_argument("name", help=sat_api_help("remove"), nargs="+")
        remove_parser.add_argument("-y", "--yes", help=sat_api_help("yes"), action="store_true")
        remove_parser.set_defaults(func=removeSatAPI)

        clear_parser = subparsers.add_parser("clear")
        clear_parser.add_argument("-y", "--yes", help=sat_api_help("yes"), action="store_true")
        clear_parser.set_defaults(func=clearSatAPI)

        # -*- Set base output verbosity level -*-
        Stream.set_lvl(1)

        if argv is None:
            args: argparse.Namespace = parser.parse_args()
        else:
            args: argparse.Namespace = parser.parse_args(argv)

        # -*- Load Interfaces -*-
        SatAPI._load()

        # Exhibits Compiler Command line arguments
        stdlog[3] << f'Command line args:\n{";".join(f"{k}={v!r}" for k, v in vars(args).items())}'

        if args.func is None:
            stdout[1] << "Available Interfaces:"
            for (i, option) in enumerate(SatAPI.options(), start=1):
                stdout[1] << f"{i}. {option}"
        else:
            args.func(args)
        
        exit(EXIT_SUCCESS)
        
__all__ = ["SatAPICLI"]



"""
"""

# -*- Satyrus -*-
from satyrus import SatAPI, Posiform

# Third-Party
import gurobipy as gp
from cstream import devnull

class gurobi(SatAPI):
    """"""

    def solve(self, energy: Posiform, **params: dict) -> tuple[dict, float]:
        """"""
        
        # Retrieve QUBO
        x, Q, c = energy.qubo()

        try:
            # pylint: disable=no-member
            
            # Create a new model
            model = gp.Model("MIP")

            X = model.addMVar(len(x), vtype=gp.GRB.BINARY, name="")

            model.setMObjective(Q, None, 0.0, X, X, None, sense=gp.GRB.MINIMIZE)

            with devnull:
                model.optimize()

            y = list(model.getVars())

            s = {k: int(y[i].x) for k, i in x.items()}
            e = model.objVal
        except gp.GurobiError as exc:
            self.error(f"(Gurobi code {exc.errno}) {exc}")
        else:
            return (s, e + c)



"""
"""
# Future Imports
from __future__ import annotations

# Standard Library
import os
import posixpath
import traceback
from functools import wraps, reduce

os.path.join = posixpath.join

def arange(start: object, stop: object = None, step: object = None):
    """\
    arange(stop) -> [0, 1, ..., stop]
    arange(start, stop) -> [start, start + 1, ..., stop]
    arange(start, stop, step) -> [start, start + step, ..., stop]
    """
    ## Value Checking
    if step == 0:
        raise ValueError("Step must be non-zero.")

    ## Defaults
    if stop is None:
        stop = start
        start = 0

    if step is None:
        if start > stop:
            step = -1
        elif start <= stop:
            step = 1

    ## A bit more Value Checking
    if start > stop and step > 0 or start < stop and step < 0:
        raise ValueError(f"Infinite range in [{start}:{stop}:{step}].")

    ## Type Coercing
    if any(type(s) is float for s in {start, stop, step}):
        x = float(start)
        stop = float(stop)
        step = float(step)
    elif any(type(s) is int for s in {start, stop, step}):
        x = int(start)
        stop = int(stop)
        step = int(step)
    else:
        x = start

    ## Iterator Loop
    if step > 0:
        while x <= stop:
            yield x
            x += step
    else:
        while x >= stop:
            yield x
            x += step


def log(target: str = "satyrus.log") -> str:
    """\
    """
    trace: str = traceback.format_exc()
    with open(target, "w", encoding="utf8") as file:
        file.write(trace)
    return trace

def prompt(question: str, answers: dict[str, object]) -> Any:
    while True:
        answer: str = input(question).strip()

        if answer not in answers:
            continue
        else:
            return answers[answer]

__all__ = ["arange", "log", "prompt"]



"""
"""
# Future Import
from __future__ import annotations

# Third-Party
from cstream import stderr, stdlog, stdwar

# Local
from .memory import Memory
from ..satlib import Source, Stack, Posiform, Timing
from ..parser import SatParser
from ..parser.legacy import SatLegacyParser
from ..error import (
    SatTypeError,
    SatReferenceError,
    SatError,
    SatExit,
    SatWarning,
)
from ..types import Expr, SatType, Number, Var, Array
from ..symbols import OPT, CMD_INIT, CMD_SCRIPT
from ..symbols import CONS_INT, CONS_OPT

class SatCompiler:
    def __init__(self, instructions: dict, parser: SatParser = None, env: dict = None):
        """
        Parameters
        ----------
        instructions : dict
                Dictionary containing the compiler's instruction set. Matches string command identifiers with respective functions.
        parser : SatParser
                Allows for chosing alternative parsers for a given compiler. Main uses are for parsing multiple versions of an evolving language.
        env : dict
                Initial compiler environment configuration.
        """
        ## Build Instruction Set
        if type(instructions) is not dict:
            raise TypeError("'instructions' must be of type 'dict'.")
        elif not instructions:
            raise ValueError("Empty Instruction Set.")
        else:
            self.instructions = instructions

        ## Initialize parser
        if parser is None:
            self.parser = SatParser()
        elif type(parser) is not SatParser and type(parser) is not SatLegacyParser:
            raise TypeError(f"'parser' must be of type 'SatParser' or 'SatLegacyParser, not {type(parser)}'.")
        else:
            self.parser = parser

        # -*- Flags -*-
        self.__flags__ = {"slow": False}

        # Source Code
        self.source: Source = None

        # Compiler Memory
        self.memory = Memory()

        # Compiler Environment Variables
        if env is None:
            self.env = {}
        else:
            self.env = dict(env)

        # Constraints & Penalty Levels
        self.constraints = {CONS_INT: [], CONS_OPT: []}
        self.penalties = {}

        # Output
        self.energy = Posiform()

        # ? Exit code ?
        self.code = 0

        # Errors
        self.error_stack = Stack()

    def __enter__(self, *args, **kwargs):
        ## Instantiate subcompiler
        subcomp = self.__class__(self.instructions, self.parser)
        subcomp.memory = self.memory.copy
        subcomp.env = self.env.copy()
        return subcomp

    def __exit__(self, *args, **kwargs):
        return None

    def __lshift__(self, error: SatError) -> None:
        """Pushes error into compiler stack.

        Parameters
        ----------
        error: SatError
                Error to be raised in the next 'compiler.checkpoint()' call.
        """
        self.error_stack.push(error)

    def __lt__(self, warning: SatWarning) -> None:
        """Prints Warning to stdout.

        Parameters
        ----------
        warning: SatWarning
        """
        stdwar[1] << warning

    def __getitem__(self, opt_level: int) -> bool:
        """Optimization Level check.

        Parameters
        ----------
        opt_level: int
                Desired optimization level.

        Returns
        -------
        bool
                Equivalent to (compiler.opt >= opt_level)

        Example
        -------
        This requires optimization level to be at least 'n' to execute block.
        >>> if compiler[n]:
        ...     block
        """
        return self.env[OPT] >= opt_level

    def __call__(self, energy: Posiform | None):
        self.energy = energy

    def flag(self, key: str, value: bool = None) -> bool:
        if value is None:
            if key in self.__flags__:
                return self.__flags__[key]
            else:
                return False
        else:
            self.__flags__[key] = value
            return value

    @Timing.timer(level=2, section="Compiler.compile")
    def compile(self, source: Source) -> int:
        """
        Parameters
        ----------
        source: Source
                String-like object created from '.sat' source-code file.

        Returns
        -------
        int
                Compiler exit code.
        """
        try:
            ## Clear compiler state
            self.code = 0
            self.source = source

            BYTECODE = self.parse(source)

            ## Adds special instructions RUN_INIT, RUN_SCRIPT in both ends
            self.execute([CMD_INIT, *BYTECODE, CMD_SCRIPT])
        except SatExit as error:
            self.code = error.code
            self.energy = self.source = None
            return self.code
        except SatError as error:
            stderr[0] << error
            self.energy = self.source = None
            self.code = error.code
            return self.code
        except Exception as error:
            self.energy = self.source = None
            self.code = 1
            raise error
        else:
            return self.code

    def parse(self, source: Source, load: bool = False) -> list:
        """Parses source into bytecode.

        Parameters
        ----------
        source : Source
                String-like object created from '.sat' source-code file.

        Returns
        -------
        list
                Bytecode sequence.
        """
        ## Parse code into bytecode
        return self.parser.parse(source)

    def execute(self, bytecode: list):
        """Executes bytecode sequentially.

        Parameters
        ----------
        bytecode : list
                Sequence of tuples (CMD, *args) to be executed.
        """

        for stmt in bytecode:
            stdlog[3] << "\n\t".join([f"CMD {stmt[0]}", *(f"{i} {x}" for i, x in enumerate(stmt[1:]))])
            self.exec(stmt)
        else:
            stdlog[3] << ""

        self.checkpoint()

    def exit(self, code: int = 0):
        """Exits compiler session with a given code.

        Parameters
        ----------
        code : int
                Compiler exit code.
        """
        raise SatExit(code)

    def exec(self, stmt: tuple):
        """Executes a single bytecode instruction.

        Parameters
        ----------
        stmt : tuple
                (CMD, *args) pair.
        """
        cmd, *args = stmt
        return self.instructions[cmd](self, *args)

    def __contains__(self, item: Var):
        """ """
        return item in self.memory

    def memset(self, name: Var, value: SatType):
        """ """
        try:
            return self.memory.memset(name, value)
        except SatTypeError as error:
            self << error
        finally:
            self.checkpoint()

    def memget(self, name: Var) -> SatType:
        """

        Returns
        -------
        SatType
                Internal Satyrus Object retrieved from compiler memory.
        """
        try:
            return self.memory.memget(name)
        except SatReferenceError as error:
            self << error
        finally:
            self.checkpoint()

    def evaluate(
        self,
        item: SatType,
        miss: bool = True,
        calc: bool = True,
        null: bool = False,
        track: bool = False,
        context: dict = None,
    ) -> SatType:
        """
        Parameters
        ----------
        item : SatType
                Internal Satyrus object to be evaluated.
        miss : bool
                If True, raises SatReferenceError when undefined variable is found. Otherwise, returns the variable itself.
        calc : bool
                If True, applies Expr.calculate to expressions retrived by compiler evaluation.
        null : bool
                If True, allows Python null (None) to be bypassed during evaluation. Otherwise, raises (critical) ValueError.
        track : bool
                If True, attempts to propagate 'lexinfo' into evaluation results.
        context : dict
                Temporary memory scope to be pushed into memory stack before evaluation and released just after.

        Example
        -------
        ...
        >>> compiler = SatCompiler(*args, **kwargs)
        >>> n = Var('n')
        >>> compiler.memset(n, Number('0.5'))
        >>> compiler.evaluate(n)
        Number('0.5')
        >>> compiler.evaluate(Number('7.4'))
        Number('7.4')
        """
        if context is not None:
            self.push(context)
            result = self._evaluate(item, miss, calc, null)
            self.pop()
        else:
            result = self._evaluate(item, miss, calc, null)

        if track:
            self.source.propagate(item, result)

        return result

    def _evaluate(self, item: SatType, miss: bool = True, calc: bool = True, null: bool = False) -> SatType:
        """ """
        if item is None:
            if null:
                return None
            else:
                raise ValueError("Invalid None appearance in compiler 'evaluate'.")
        elif type(item) is Expr:
            if calc:
                return Expr.calculate(Expr(item.head, *(self._evaluate(p, miss, calc) for p in item.tail)))
            else:
                return Expr(item.head, *(self._evaluate(p, miss, calc) for p in item.tail))
        elif type(item) is Number:
            return item
        elif type(item) is Array:
            return item
        elif type(item) is Var:
            try:
                return self.memory.memget(item)
            except SatReferenceError as error:
                if miss:
                    self << error
                else:
                    return item
            finally:
                self.checkpoint()
        else:
            raise TypeError(f"Invalid Type '{type(item)}' for '{item!r}' in 'compiler.evaluate'.")

    def push(self, context: dict = None):
        """Adds new scope to the top of the memory stack. Variable definition is done by 'SatCompiler.memset' and will follow its rules for referencing.

        Parameters
        ----------
        context : dict, optional
                Mapping between variables and values.
        """
        self.memory.push()

        if context is not None:
            for k, v in context.items():
                self.memset(k, v)

    def pop(self, depth: int = 1):
        """Removes a given number of layers from the memory stack. If removal goes beyond global scope, an exception is raised.

        Parameters
        ----------
        depth: int, optional
                How many layers should be removed from the memory stack, defaults to a single one.
        """
        self.memory.pop(depth)

    def interrupt(self):
        """Interrupts compilation after outputing error messages to stderr."""
        while self.error_stack:
            error = self.error_stack.popleft()
            stderr[0] << error
        else:
            self.exit(1)

    def checkpoint(self):
        """If there is any unhandled exception in the compiler's error stack, interrupts compilation via a 'SatCompiler.interrupt' call"""
        if self.status:
            self.interrupt()

    @property
    def status(self) -> bool:
        """Evaluates to True if there is any unhandled exception in the compiler's error stack. Otherwise, returns False.

        Returns
        -------
        bool
        """
        return bool(self.error_stack)



from os import urandom
import sys
import site
from pathlib import Path

PACKAGE = "satyrus_data"


def package_path(*, package: str = PACKAGE, fname: str = None) -> Path:
    """"""
    if fname is None:
        sys_path = Path(sys.prefix).joinpath(package)
        usr_path = Path(site.USER_BASE).joinpath(package)
    else:
        sys_path = Path(sys.prefix).joinpath(package, fname)
        usr_path = Path(site.USER_BASE).joinpath(package, fname)

    if not sys_path.exists():
        if not usr_path.exists():
            raise FileNotFoundError(f"File or Directory '{fname}' not installed in '{package}'")
        else:
            return usr_path
    else:
        return sys_path


__all__ = ["PACKAGE", "package_path"]



""""""
# Local
from .base import SatType
from .expr import Expr
from .number import Number
from .var import Var
from ..error import SatIndexError
from ..symbols import T_IDX

class Array(SatType):
    """ :: Array ::
        ===========
    """
    def __init__(self, var: Var, shape: tuple):
        SatType.__init__(self)
        
        self.array = {}

        self.var = var
        self.shape = shape
        self.dim = len(shape)

    def _IDX_(self, idx: tuple):
        if len(idx) > self.dim:
            raise SatIndexError(f"Too much indices for {self.dim}-dimensional array.", target=idx[-1])
        elif len(idx) == 1:
            return self[idx[0]]
        else:
            i, *idx = idx
            return self[i]._IDX_(idx)

    def __getitem__(self, i: Number):
        if type(i) is Number and i.is_int:
            j = int(i)
            if j in self.array:
                return self.array[j]
            else:
                n = int(self.shape[0])
                if 1 <= j <= n:
                    if len(self.shape) > 1:
                        self.array[j] = Array(self.var._IDX_((i,)), self.shape[1:])
                        return self.array[j]
                    else:
                        return self.var._IDX_((i,))
                else:
                    raise SatIndexError(f"Index {i} is out of bounds [1, {n}].", target=i)
        elif type(i) is Var:
            return Expr(T_IDX, self, i)
        else:
            raise SatIndexError(f"Invalid index `{i}` of type `{type(i)}`.", target=i)

    def __setitem__(self, i: Number, value: SatType):
        if type(i) is Number and i.is_int:
            if self.dim > 1:
                raise SatIndexError(f"Attempt to assign to {self.dim}-dimensional array.", target=i)
            else:
                j = int(i)
                n = int(self.shape[0])
                if 1 <= j <= n:
                    self.array[j] = value
                else:
                    raise SatIndexError(f"Index {i} is out of bounds [1, {n}].", target=i)                 
        else:
            raise SatIndexError(f"Invalid index {i}.", target=i)

    def __str__(self):
        return f"{self.var}" #+ "".join([f"[{n}]" for n in self.shape])

    def __repr__(self):
        return f"Array({self.var!r}, {self.shape!r})"

    def __hash__(self):
        return hash(sum(hash((k, hash(v))) for k, v in self.array.items()))
        
    @classmethod
    def from_buffer(cls, name, shape, buffer: dict):
        array = cls(name, shape)
        for idx in buffer:
            subarray = array
            for i in idx[:-1]:
                subarray = subarray[Number(i)]
            subarray[Number(idx[-1])] = buffer[idx]
        return array

    @classmethod
    def from_dict(cls, name: Var, shape: tuple, buffer: dict):
        """
        """
        return cls.from_buffer(name, shape, buffer)

    @classmethod
    def from_list(cls, name: Var, shape: tuple, buffer: dict):
        """
        """
        return cls.from_buffer(name, shape, { idx : val for (idx, val) in buffer })

    @property
    def is_array(self) -> bool:
        return True


from .base import SatType
from ..satlib import Source

class String(SatType, str):
    """ 
    Parameters
    ----------
    buffer : str
    """
    def __new__(cls, buffer: str, source: Source = None, lexpos: int = None):
        return str.__new__(cls, buffer)

    def __init__(self, buffer: str, source: Source = None, lexpos: int = None):
        SatType.__init__(self, source=source, lexpos=lexpos)


""" DEF_CONSTANT
    ============

    STATUS: OK
"""

from ..compiler import SatCompiler
from ...error import SatReferenceError
from ...types import SatType, Var

def def_constant(compiler: SatCompiler, name : Var, value : SatType):
    """ DEF_CONSTANT
        ============
    """
    compiler.memset(name, compiler.evaluate(value, miss=True, calc=True, null=False, track=True))


# Standard Library
from time import perf_counter as clock
from functools import wraps

# Third-Party
from cstream import stdlog


class Timing(object):
    """"""

    class Timer(object):
        """"""
        BAR = "■"
        BARS = 24
        EMPTY = " "

        def __init__(self, prec=2, *, bar: str = None, arrow: str = None):
            self.pattern = f"Time elapsed @ {{}}:  {{:.{prec}f}}s"
            self.sections = {}

        def __call__(self, level: int = 0, section: str = None):
            if type(level) is int or section is not None:

                def decor(callback: callable) -> callable:
                    @wraps(callback)
                    def timed_callback(*args, **kwargs):
                        t = clock()
                        _ = callback(*args, **kwargs)
                        T = clock() - t
                        self.sections[section] = T
                        stdlog[level] << self.pattern.format(section, T)
                        return _

                    return timed_callback

                return decor
            elif callable(level):
                callback = level

                @wraps(callback)
                def timed_callback(*args, **kwargs):
                    t = clock()
                    _ = callback(*args, **kwargs)
                    T = clock() - t
                    self.sections[section] = T
                    stdlog[level] << self.pattern.format(section, T)
                    return _

                return timed_callback
            else:
                raise TypeError("Place some info here.")

        def reset(self):
            self.sections.clear()

        def report(self):
            return [(k, t) for k, t in self.sections.items() if k is not None]

        def show_report(self, level: int = 0):
            R = self.report()
            L = max(len(k) for k, _ in R)
            T = sum(t for _, t in R)
            stdlog[level] << f"Time elapsed: {T:.2f}s"
            for k, t in R:
                stdlog[level] << f"{self.__fill(k, L)}\t{self.__bar(t, T)}"

        @classmethod
        def __fill(cls, k: str, L: int):
            l = len(k)
            return k + " " * (L - l)

        @classmethod
        def __bar(cls, t: float, T: float):
            x = (t / T) if T else 0.0
            return f"[{cls.__full(x)}{cls.__empty(x)}]{t:6.2f}s ({x * 100.0:6.2f}%)"

        @classmethod
        def __bars(cls, x: float) -> int:
            return int(cls.BARS * x)

        @classmethod
        def __full(cls, x: float) -> str:
            return cls.BAR * cls.__bars(x)

        @classmethod
        def __empty(cls, x: float) -> str:
            return cls.EMPTY * (cls.BARS - cls.__bars(x))

    timer = Timer()


__all__ = ["Timing"]



"""
"""
# Third-Party
from ply import lex, yacc
from cstream import stderr

# Local
from ...satlib import Source, Stack
from ...error import (
    SatParserError,
    SatSyntaxError,
    SatWarning,
    SatExit,
)
from ...types import Expr, Var, Number, String, SatType
from ...symbols import (
    DEF_CONSTANT,
    DEF_ARRAY,
    DEF_CONSTRAINT,
    CONS_INT,
    CONS_OPT,
    T_ADD,
    T_NEG,
    T_IDX,
    T_FORALL,
    T_EXISTS,
)


def regex(pattern: str):
    def decor(callback):
        callback.__doc__ = pattern
        return callback

    return decor


class SatLegacyLexer(object):

    ## Keywords
    keywords = [
        ## load
        "from",
        ## logic
        "and",
        "or",
        "not",
        "xor",
        ## constraints
        "optgroup",
        "intgroup",
        "forall",
        "exists",
        "in",
        "where",
        ## penalties
        "correlation",
        "penalties",
        "level",
    ]

    reserved = {kw: kw.upper() for kw in keywords}

    ## List of token names.
    tokens = (
        ## 'FORALL', 'EXISTS', # 'EXISTS_ONE', # quantifiers
        ## 'NOT', 'AND', 'OR', 'XOR', # logical
        "IMP",
        "RIMP",
        "IFF",  # extended logical
        "ENDL",
        "COMMA",
        "ASSIGN",
        "EQ",  # equal
        "GT",  # greater than
        "LT",  # less than
        "GE",  # greater or equal
        "LE",  # less or equal
        "NE",  # not equal
        "MUL",
        "DIV",
        "MOD",
        "ADD",
        "SUB",  # arithmetic
        "DOTS",
        "LBRA",
        "RBRA",
        "LPAR",
        "RPAR",
        "LCUR",
        "RCUR",
        "NAME",
        "NUMBER",
        "STRING",
    ) + tuple(map(str.upper, keywords))

    def __init__(self, source, **kwargs):
        self.lexer = lex.lex(object=self, **kwargs)
        self.source = source

        self.error_stack = Stack()

    def __lshift__(self, error: SatParserError):
        self.error_stack.push(error)

    def exit(self, code: int):
        raise SatExit(code)

    def interrupt(self):
        """ """
        while self.error_stack:
            error = self.error_stack.popleft()
            stderr[0] << error
        else:
            self.exit(1)

    def checkpoint(self):
        """ """
        if self.error_stack:
            self.interrupt()

    # Regular expression rules for tokens
    t_DOTS = r"\:"

    t_LBRA = r"\["
    t_RBRA = r"\]"

    t_LPAR = r"\("
    t_RPAR = r"\)"

    t_LCUR = r"\{"
    t_RCUR = r"\}"

    t_IMP = r"\-\>"

    t_RIMP = r"\<\-"

    t_IFF = r"\<\-\>"

    t_COMMA = r"\,"

    t_DIV = r"\/"

    t_MOD = r"\%"

    t_MUL = r"\*"

    t_ADD = r"\+"

    t_SUB = r"\-"

    t_ENDL = r"\;"

    t_ASSIGN = r"\="

    t_EQ = r"\=\="
    t_GT = r"\>"
    t_LT = r"\<"
    t_GE = r"\>\="
    t_LE = r"\<\="
    t_NE = r"\!\="

    @regex(r"\"(\\.|[^\"])*\"")
    def t_STRING(self, t):
        t.value = String(t.value[1:-1], source=self.source, lexpos=t.lexpos)
        return t

    @regex(r"\n")
    def t_newline(self, t):
        self.lexer.lineno += 1
        return None

    @regex(r"[a-zA-Z_][a-zA-Z0-9_]*")
    def t_NAME(self, t):
        t.type = self.reserved.get(str(t.value), "NAME")
        t.value = String(t.value, source=self.source, lexpos=t.lexpos)
        return t

    @regex(r"[0-9]*\.?[0-9]+([Ee][-+]?[0-9]+)?")
    def t_NUMBER(self, t):
        t.value = Number(t.value, source=self.source, lexpos=t.lexpos)
        return t

    # String containing ignored characters between tokens
    t_ignore = " \t"
    t_ignore_COMMENT = r"\/\/.*"

    @regex(r"\#\*[\s\S]*?\*\/")
    def t_ignore_MULTICOMMENT(self, t):
        self.lexer.lineno += str(t.value).count("\n")
        return None

    def t_error(self, t):
        stderr << f"Unknown token '{t.value}' at line {t.lineno}"
        if t:
            stderr << f"Syntax Error at line {t.lineno}:"
            stderr << self.source.lines[t.lineno - 1]
            stderr << f'{" " * (self.chrpos(t.lineno, t.lexpos))}^'
        else:
            stderr << "Unexpected End Of File."

    def chrpos(self, lineno, lexpos):
        return lexpos - self.source.table[lineno - 1] + 1


class SatLegacyParser(object):

    tokens = SatLegacyLexer.tokens

    extra_keywords = [
        ## load
        "from",
        ## constraints
        "optgroup",
        "intgroup",
        "forall",
        "exists",
        "in",
        "where",
        ## penalties
        "correlation",
        "penalties",
        "level",
    ]

    precedence = (
        ("left", "ADD", "SUB"),
        ("left", "MUL", "DIV"),
        ("left", "XOR", "OR"),
        ("left", "AND"),
        ("left", "NOT"),
        ("left", "IMP", "RIMP", "IFF"),
        ("left", "EQ", "NE", "GE", "LE", "GT", "LT"),
        ("left", "DOTS"),
        ("left", "LCUR", "RCUR"),
        ("left", "LPAR", "RPAR"),
        ("left", "NAME"),
        ("left", "STRING", "NUMBER"),
        ("left", "LBRA", "RBRA"),
        ("left", "ENDL"),
    )

    SYMBOL_TABLE = {"and": "&", "or": "|", "xor": "^", "not": "~"}

    def symbol_table(self, symbol: str):
        if symbol in self.SYMBOL_TABLE:
            return self.SYMBOL_TABLE[symbol]
        else:
            return symbol

    def __init__(self):
        ## Input
        self.source = None
        ## Output
        self.bytecode = None

        ## Lex & Yacc
        self.lexer = None
        self.parser = None

        self.error_stack = Stack()

        ## Lazy definitions
        self.array_table = {}
        self.constraint_table = {}

    def __lshift__(self, error: SatParserError):
        self.error_stack.push(error)

    def exit(self, code: int):
        raise SatExit(code)

    def interrupt(self):
        """ """
        while self.error_stack:
            error = self.error_stack.popleft()
            stderr[0] << error
        else:
            self.exit(1)

    def checkpoint(self):
        """ """
        if self.error_stack:
            self.interrupt()

    def parse(self, source: Source):
        ## Input
        self.source = source

        ## Initialize Lexer
        self.lexer = SatLegacyLexer(self.source)

        ## Initialize Parser
        self.parser = yacc.yacc(module=self, write_tables=False, debug=False)

        ## Run Parser
        if not self.source:
            self << SatSyntaxError("Empty File.", target=self.source.eof)
        else:
            self.parser.parse(self.source)

        ## Checkpoint
        self.checkpoint()

        ## Output
        return self.bytecode

    def run(self, code: list):
        self.bytecode = [cmd for cmd in code if cmd is not None]

    def p_start(self, p):
        """start : definitions constraints penalties"""
        self.run([*p[1], *p[2], *p[3]])

    def p_definitions(self, p):
        """definitions : definitions definition
        | definition
        """
        if len(p) == 3:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_definition(self, p):
        """definition : def_constant ENDL
        | dec_array ENDL
        | def_array ENDL
        | FROM STRING ENDL
        """
        p[0] = p[1]

    def p_constraints(self, p):
        """constraints : constraints constraint
        | constraint
        """
        if len(p) == 3:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_constraint(self, p):
        """constraint : dec_constraint ENDL"""
        p[0] = p[1]

    def p_def_constant(self, p):
        """def_constant : varname ASSIGN constant"""
        name = p[1]
        value = p[3]
        p[0] = (DEF_CONSTANT, name, value)

    def p_literal(self, p):
        """literal : constant
        | varname
        """
        p[0] = p[1]

    def p_constant(self, p):
        """constant : NUMBER"""
        p[0] = p[1]

    def p_varname(self, p):
        """varname : NAME"""
        p[0] = Var(p[1], source=self.source, lexpos=p.lexpos(1))

    def p_dec_array(self, p):
        """dec_array : varname LPAR shape RPAR"""
        name = p[1]
        shape = p[3]

        ## declare array
        self.array_table[name] = shape

        p[0] = (DEF_ARRAY, name, shape, None)

    def p_def_array(self, p):
        """def_array : varname ASSIGN array_buffer"""
        name = p[1]
        buffer = p[3]

        if name not in self.array_table:
            self << SatSyntaxError("Assignment to undeclared array structure.")
            self.checkpoint()
        else:
            shape = self.array_table[name]
            p[0] = (DEF_ARRAY, name, shape, buffer)

    def p_shape(self, p):
        """shape : shape COMMA expr
        | expr
        """
        if len(p) == 4:
            p[0] = (*p[1], p[3])
        else:
            p[0] = (p[1],)

    def p_array_buffer(self, p):
        """array_buffer : LBRA array RBRA"""
        p[0] = p[2]

    def p_array(self, p):
        """array : array ENDL array_item
        | array_item
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [
                p[1],
            ]

    def p_array_item(self, p):
        """array_item : array_index DOTS expr"""
        p[0] = (p[1], p[3])

    def p_array_index(self, p):
        """array_index : literal_seq"""
        p[0] = p[1]

    def p_literal_seq(self, p):
        """literal_seq : literal_seq COMMA literal
        | literal
        """
        if len(p) == 4:
            p[0] = (*p[1], p[3])
        else:
            p[0] = (p[1],)

    def p_dec_constraint_loops(self, p):
        """dec_constraint : group varname DOTS loops DOTS expr"""
        type_ = p[1]
        name = p[2]

        loops = p[4]
        expr = p[6]

        constraint = (type_, name, loops, expr)

        if name in self.constraint_table:
            self.constraint_table[name].append(constraint)
        else:
            self.constraint_table[name] = [constraint]

    def p_dec_constraint(self, p):
        """dec_constraint : group varname DOTS expr"""
        type_ = p[1]
        name = p[2]

        expr = p[4]

        constraint = (type_, name, [], expr)

        if name in self.constraint_table:
            self.constraint_table[name].append(constraint)
        else:
            self.constraint_table[name] = [constraint]

    def p_group(self, p):
        """group : INTGROUP
        | OPTGROUP
        """
        group_name = p[1]
        if str(group_name) == "intgroup":
            name = String(CONS_INT)
        elif str(group_name) == "optgroup":
            name = String(CONS_OPT)
        else:
            raise NameError(f"Invalid group {group_name}")

        p[0] = self.source.propagate(group_name, name, out=True)

    def p_loops(self, p):
        """loops : loops ENDL loop_stack
        | loop_stack
        """
        if len(p) == 4:
            p[0] = [*p[1], *p[3]]
        elif len(p) == 2:
            p[0] = p[1]
        elif len(p) == 1:
            p[0] = []

    def p_loop(self, p):
        """loop_stack : quant LCUR var_stack RCUR WHERE domains AND conditions
        | quant LCUR var_stack RCUR WHERE domains
        """
        loop_stack = []

        var_stack = set(p[3])

        domains = p[6]

        if len(p) == 9:
            conditions = p[8]
        else:
            conditions = None

        stack = set()

        quant = {"forall": T_FORALL, "exists": T_EXISTS}[p[1].lower()]

        for var, a, b in domains:
            stack.add(var)

            loop = [quant, var, (a, b, None), None]

            loop_stack.append(loop)

        if var_stack != stack:
            self << SatSyntaxError(
                "Loop and domain mismatch.", target=(var_stack ^ stack).pop()
            )
            self.checkpoint()

        loop_stack[-1][-1] = conditions

        p[0] = loop_stack

    def p_var_stack(self, p):
        """var_stack : var_stack COMMA varname
        | varname
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [
                p[1],
            ]

    def p_domains(self, p):
        """domains : domains COMMA domain
        | domain
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [
                p[1],
            ]

    def p_quant(self, p):
        """quant : FORALL
        | EXISTS
        """
        p[0] = p[1]

    def p_domain(self, p):
        """domain : varname IN LPAR expr COMMA expr RPAR"""
        p[0] = (p[1], p[4], p[6])

    def p_conditions(self, p):
        """conditions : conditions COMMA condition
        | condition
        """
        if len(p) == 4:
            p[0] = [*p[1], p[3]]
        else:
            p[0] = [
                p[1],
            ]

    def p_condition(self, p):
        """condition : expr EQ expr
        | expr GT expr
        | expr LT expr
        | expr GE expr
        | expr LE expr
        | expr NE expr
        """
        p[0] = Expr(self.symbol_table(p[2]), p[1], p[3])

    def p_condition_expr(self, p):
        """condition : expr"""
        p[0] = p[1]

    def p_expr(self, p):
        """expr : literal"""
        p[0] = p[1]

    def p_unary_expr(self, p):
        """expr : NOT expr
        | ADD expr
        | SUB expr
        """
        p[0] = Expr(self.symbol_table(p[1]), p[2])

    def p_binary_expr(self, p):
        """expr : expr AND expr
        | expr OR expr
        | expr XOR expr
        | expr ADD expr
        | expr SUB expr
        | expr MUL expr
        | expr DIV expr
        | expr MOD expr
        | expr IMP expr
        | expr RIMP expr
        | expr IFF expr
        """
        symbol = self.symbol_table(p[2])
        if symbol == T_NEG:
            p[0] = Expr(T_ADD, p[1], Expr(T_NEG, p[3]))
        else:
            p[0] = Expr(symbol, p[1], p[3])

    def p_expr_index(self, p):
        """expr : expr LBRA expr RBRA"""
        p[0] = Expr(T_IDX, p[1], p[3])

    def p_expr_par(self, p):
        """expr : LPAR expr RPAR"""
        p[0] = p[2]

    def p_penalties(self, p):
        """penalties : PENALTIES DOTS penalty_list"""
        penalties = p[3]

        commands = []

        for name, level, cor in penalties:  # pylint: disable=unused-variable
            if name not in self.constraint_table:
                self < SatWarning(
                    f"Undefined constraint group {name} to be ignored.", target=name
                )
            else:
                index = 0
                for type_, name, loops, expr in self.constraint_table[name]:
                    if str(type_) == CONS_OPT and level == Number("0"):
                        level = None
                    commands.append(
                        (DEF_CONSTRAINT, type_, f"{name}-{index}", loops, expr, level)
                    )
                    index += 1

        p[0] = commands

    def p_penalty_list(self, p):
        """penalty_list : penalty_list penalty
        | penalty
        """
        if len(p) == 3:
            p[0] = [*p[1], p[2]]
        else:
            p[0] = [
                p[1],
            ]

    def p_penalty(self, p):
        """penalty : varname DOTS LEVEL NUMBER COMMA CORRELATION NUMBER ENDL
        | varname DOTS LEVEL NUMBER ENDL
        """
        if len(p) == 9:
            p[0] = (p[1], p[4], p[7])
        else:
            p[0] = (p[1], p[4], None)

    def p_error(self, t):
        stderr[5] << f"Error Token: `{t}`"
        target = SatType()
        if t:
            target.lexinfo = {
                "lineno": t.lineno,
                "lexpos": t.lexpos,
                "chrpos": self.chrpos(t.lineno, t.lexpos),
                "source": self.source,
            }
            msg = "Invalid Syntax"
        else:
            lineno = len(self.source.lines) - 1
            lexpos = len(self.source.lines[lineno]) - 1
            target.lexinfo = {
                "lineno": lineno,
                "lexpos": lexpos,
                "chrpos": self.chrpos(lineno, lexpos),
                "source": self.source,
            }
            msg = "Unexpected End Of File."
        self << SatSyntaxError(msg=msg, target=target)
        return None

    def chrpos(self, lineno: int, lexpos: int):
        return lexpos - self.source.table[lineno - 1] + 1



from .base import SatType
from .expr import Expr
from ..error import SatValueError
from ..satlib import Source
from ..symbols import T_IDX

# pylint:disable=no-self-argument
class Var(str, SatType):
    """ """

    def __new__(cls, name: str, *, source: Source = None, lexpos: int = None):
        return str.__new__(cls, name)

    def __init__(self, name: str, *, source: Source = None, lexpos: int = None):
        SatType.__init__(self, source=source, lexpos=lexpos)

    def __repr__(self):
        return f"{self.__class__.__name__}({str.__repr__(self)})"

    def _IDX_(s, i: tuple):
        r = str(s)
        while i:
            j, *i = i
            if j.is_number:
                if j > 0 and j.is_int:
                    r = f"{r}_{j}"
                else:
                    raise SatValueError(
                        f"Index must be a positive integer, not '{j}'.", target=j
                    )
            else:
                return Expr(T_IDX, Var(r), j, *i)
        else:
            return Var(r)

    @property
    def is_var(self) -> bool:
        return True

    @property
    def is_expr(self) -> bool:
        return False

    @property
    def is_array(self) -> bool:
        return False

    @property
    def is_number(self) -> bool:
        return False



from satyrus import SatAPI, Posiform

class dwave(SatAPI):
    """"""

    def solve(self, energy: Posiform, *, num_reads=1_000, num_sweeps=1_000, **params: dict) -> tuple[dict, float]:
        """
        Parameters
        ----------
        energy : Posiform
            Input Expression
        params : dict (optional)
            num_reads : int = 1_000
            num_sweeps : int = 1_000
        """
        import neal

        # Parameter check
        self.check_params(num_reads=num_reads, num_sweeps=num_sweeps)

        self.warn("D-Wave option currently running local simulated annealing since remote connection to quantum host is unavailable")

        self.log(f"Dwave Parameters: num_reads={num_reads}; num_sweeps={num_sweeps}")

        sampler = neal.SimulatedAnnealingSampler()

        x, Q, c = energy.qubo()

        sampleset = sampler.sample_qubo(Q, num_reads=num_reads, num_sweeps=num_sweeps)

        y, e, *_ = sampleset.record[0]

        s = {k: int(y[i]) for k, i in x.items()}

        return (s, e + c)

    def check_params(self, **params: dict):
        if "num_reads" in params:
            num_reads = params["num_reads"]
            if not isinstance(num_reads, int) or num_reads <= 0:
                self.error("Parameter 'num_reads' must be a positive integer")
        else:
            self.error("Missing parameter 'num_reads'")

        if "num_sweeps" in params:
            num_sweeps = params["num_sweeps"]
            if not isinstance(num_sweeps, int) or num_sweeps <= 0:
                self.error("Parameter 'num_sweeps' must be a positive integer")
        else:
            self.error("Missing parameter 'num_sweeps'")


from .array import Array
from .base import SatType
from .expr import Expr
from .python import PythonObject
from .number import Number
from .string import String
from .var import Var

__all__ = ["Array", "Var", "SatType", "Expr", "Number", "String", "PythonObject"]



"""
"""
# Typing
from __future__ import annotations
from typing import Callable

# Standard Library
import itertools as it
from functools import reduce

# Local
from .base import SatType, MetaSatType
from .number import Number
from ..satlib import Source
from ..symbols import (
    T_IDX,
    T_AND,
    T_OR,
    T_ADD,
    T_MUL,
    T_POW,
    T_IFF,
    T_IMP,
    T_RIMP,
    T_NOT,
    T_NEG,
    T_XOR,
    T_DIV,
    T_MOD,
    T_EQ,
    T_NE,
    T_GE,
    T_LE,
    T_GT,
    T_LT,
    # Operand types
    T_LOGICAL,
    T_ARITHMETIC,
    T_EXTRA,
)


class Expr(tuple, SatType, metaclass=MetaSatType):
    """
    This is one of the core elements of the system. This is used to represent ASTs.
    """

    # -*- Object Creation -*-
    def __new__(
        cls,
        head: str,
        *tail: tuple,
        sort: bool = True,
        flat: bool = True,
        source: Source = None,
        lexpos: int = None,
    ):
        """An Expr instance is a (head, tail) pair representing an expression. Here is applied, according to `sort` and `flat` parameters, child sorting and n-ary expression flatening.

        Parameters
        ----------
        head : str
            Operation identifier
        tail : tuple
            Operand sequence
        """
        if sort and flat:
            return tuple.__new__(cls, (head, *cls.sort(head, cls.flat(head, tail))))
        elif sort:
            return tuple.__new__(cls, (head, *cls.sort(head, tail)))
        elif flat:
            return tuple.__new__(cls, (head, *cls.flat(head, tail)))
        else:
            return tuple.__new__(cls, (head, *tail))

    def __init__(
        self,
        head: str,
        *tail: tuple,
        sort: bool = True,
        flat: bool = True,
        source: Source = None,
        lexpos: int = None,
    ):
        SatType.__init__(self, source=source, lexpos=lexpos)
        self._flat_ = flat
        self._sort_ = sort
        self._hash_ = None

    @property
    def head(self) -> str:
        return self[0]

    @property
    def tail(self) -> list:
        return self[1:]

    # -*- Interal Expression Management -*-
    DO_SORT = {T_AND, T_OR, T_ADD, T_MUL, T_IFF}

    @classmethod
    def sorting(cls, p: SatType):
        if p.is_number:
            return (0, p)
        elif p.is_var:
            return (1, p)
        elif p.is_array:
            return (2, p.var)
        elif p.is_expr:
            return (3, hash(p))
        else:
            raise TypeError(f"Invalid type for comparison {type(p)} of '{p}'.")

    @classmethod
    def sort(cls, head: str, tail: list) -> list:
        """Returns sorted tail according to key function defined by `cls.sorting: object -> object`."""
        if head in cls.DO_SORT:
            return sorted(
                [
                    (p if not p.is_expr else (p if p._sort_ else cls(p.head, *p.tail)))
                    for p in tail
                ],
                key=cls.sorting,
            )
        else:
            return tail

    DO_FLAT = {T_AND, T_OR, T_ADD, T_MUL, T_IDX}

    @classmethod
    def flat(cls, head: str, tail: list) -> list:
        if head in cls.DO_FLAT:
            flattail = []
            for p in tail:
                if p.is_expr and (p.head == head):
                    if p._flat_:
                        tail = p.tail
                    else:
                        tail = cls.flat(p.head, p.tail)
                    flattail.extend(tail)
                else:
                    flattail.append(p)
            return flattail
        else:
            return tail

    # -*- Parenthesis -*-
    PARENTHESIS = {  ## precedence groups
        ## 1
        T_IMP: 1,
        T_RIMP: 1,
        T_IFF: 1,
        ## 2
        T_OR: 2,
        ## 3
        T_AND: 3,
        ## 4
        T_NOT: 4,
        T_ADD: 5,
        T_NEG: 4.5,
        T_MUL: 6,
    }

    @classmethod
    def parenthesise(cls, root: object, child: object) -> bool:
        if root.is_expr and child.is_expr:
            return cls._parenthesise(root.head, child.head)
        else:
            return False

    @classmethod
    def _parenthesise(cls, head_x: str, head_y: str):
        if head_x not in cls.PARENTHESIS:
            return False
        elif head_y not in cls.PARENTHESIS:
            return False
        elif cls.PARENTHESIS[head_x] > cls.PARENTHESIS[head_y]:
            return True
        else:
            return False

    # -*- Formatting -*-
    FORMAT_PATTERNS = {
        ## Logical
        T_XOR: "{1} {0} {2}",
        T_NOT: "{0}{1}",
        T_IMP: "{1} {0} {2}",
        T_RIMP: "{1} {0} {2}",
        T_IFF: "{1} {0} {2}",
        ## Arithmetic
        T_NEG: "{0}{1}",
        T_DIV: "{1} {0} {2}",
        T_MOD: "{1} {0} {2}",
        T_POW: "{1} {0} {2}",
        ## Comparison
        T_EQ: "{1} {0} {2}",
        T_LE: "{1} {0} {2}",
        T_GE: "{1} {0} {2}",
        T_LT: "{1} {0} {2}",
        T_GT: "{1} {0} {2}",
        T_NE: "{1} {0} {2}",
    }

    FORMAT_FUNCTIONS = {
        ## Arithmetic
        T_ADD: (lambda head, *tail: f" {head} ".join(tail)),
        T_MUL: (lambda head, *tail: f" {head} ".join(tail)),
        ## Logical
        T_AND: (lambda head, *tail: f" {head} ".join(tail)),
        T_OR: (lambda head, *tail: f" {head} ".join(tail)),
        ## Indexing
        T_IDX: (lambda _, *tail: f"{tail[0]}{''.join([f'[{x}]' for x in tail[1:]])}"),
    }

    def __str__(self):
        tail = [(f"({p})" if self.parenthesise(self, p) else str(p)) for p in self.tail]
        if self.head in self.FORMAT_PATTERNS:
            return self.FORMAT_PATTERNS[self.head].format(self.head, *tail)
        elif self.head in self.FORMAT_FUNCTIONS:
            return self.FORMAT_FUNCTIONS[self.head](self.head, *tail)
        elif self.head in self.FORMAT_SPECIAL:
            return self.FORMAT_SPECIAL[self.head](type(self), self.head, self.tail)
        else:
            return " ".join([self.head, *tail])

    def __repr__(self):
        return f"Expr({', '.join(map(repr, self))})"

    # -*- Hash Computing -*-
    HASH_COM = lambda h: (lambda *X: hash((h, hash(sum(map(hash, X))))))
    HASH_NCM = lambda h: (lambda *X: hash((h, hash(tuple(map(hash, X))))))
    HASH_REV = lambda h: (lambda *X: hash((h, hash(tuple(map(hash, reversed(X)))))))
    HASH_IDP = lambda h: (lambda x: (ord(h) - hash(x)))

    HASH = {
        ## Logical
        T_OR: HASH_COM(T_OR),
        T_AND: HASH_COM(T_AND),
        T_NOT: HASH_IDP(T_NOT),
        T_XOR: HASH_NCM(T_XOR),
        T_IMP: HASH_NCM(T_IMP),
        T_RIMP: HASH_REV(T_IMP),
        T_IFF: HASH_COM(T_IFF),
        ## Arithmetic
        T_ADD: HASH_COM(T_ADD),
        T_NEG: HASH_IDP(T_NEG),
        T_MUL: HASH_COM(T_MUL),
        T_DIV: HASH_NCM(T_DIV),
        T_MOD: HASH_NCM(T_MOD),
        T_POW: HASH_NCM(T_POW),
        ## Comparisons
        T_EQ: HASH_COM(T_EQ),
        T_NE: HASH_COM(T_NE),
        T_GE: HASH_NCM(T_GE),
        T_LE: HASH_NCM(T_GE),
        T_GT: HASH_NCM(T_GT),
        T_LT: HASH_NCM(T_GT),
        ## Indexing
        T_IDX: HASH_NCM(T_IDX),
    }

    def __hash__(self):
        if self._hash_ is None:
            self._hash_ = self.HASH[self.head](*self.tail)
        return self._hash_

    # -*- Computation Rules -*-
    @staticmethod
    def R_OR(cls: type, *tail: SatType) -> SatType:
        expr_table = {}
        cons_table = []

        for p in tail:
            if p.is_number:
                if p == Number("0"):
                    continue
                elif p == Number("1"):
                    return Number("1")
                else:
                    cons_table.append(p)
            else:
                expr_table[hash(p)] = p

        cons = reduce(lambda x, y: x._OR_(y), cons_table, Number("0"))

        if cons == Number("1"):
            return Number("1")
        elif cons == Number("0"):
            if len(expr_table) == 0:
                return Number("0")
            elif len(expr_table) == 1:
                return expr_table.pop(next(iter(expr_table.keys())))
            else:
                return cls(T_OR, *expr_table.values())
        else:
            if len(expr_table) == 0:
                return cons
            else:
                return cls(T_OR, cons, *expr_table.values())

    @staticmethod
    def R_AND(cls: type, *tail: SatType) -> SatType:
        expr_table = {}
        cons_table = []

        for p in tail:
            if p.is_number:
                if p == Number("1"):
                    continue
                elif p == Number("0"):
                    return Number("0")
                else:
                    cons_table.append(p)
            else:
                expr_table[hash(p)] = p

        cons = reduce(lambda x, y: x._AND_(y), cons_table, Number("1"))

        if cons == Number("0"):
            return Number("0")
        elif cons == Number("1"):
            if len(expr_table) == 0:
                return Number("1")
            elif len(expr_table) == 1:
                return expr_table.pop(next(iter(expr_table.keys())))
            else:
                return cls(T_AND, *expr_table.values())
        else:
            if len(expr_table) == 0:
                return cons
            else:
                return cls(T_AND, cons, *expr_table.values())

    @staticmethod
    def R_NOT(cls: type, x: SatType) -> SatType:
        if x.is_expr:
            if x.head == T_NOT:
                return x[1]
            elif x.head == T_AND:
                return cls(T_OR, *(p._NOT_() for p in x.tail))
            elif x.head == T_OR:
                return cls(T_AND, *(p._NOT_() for p in x.tail))
            else:
                return x._NOT_()
        else:
            return x._NOT_()

    @staticmethod
    def R_ADD(cls: type, *tail: SatType) -> SatType:
        hash_table = {}
        expr_table = {}
        cons_table = []

        for p in tail:
            if p.is_number:
                cons_table.append(p)
            elif p.is_expr and (p.head == T_MUL) and p[1].is_number:
                c, *t = p[1:]
                if len(t) == 1:
                    q = t[0]
                else:
                    q = cls(T_MUL, *t)

                if hash(q) in hash_table:
                    expr_table[hash(q)] += c
                else:
                    hash_table[hash(q)] = q
                    expr_table[hash(q)] = c
            elif hash(p) in hash_table:
                expr_table[hash(p)] += Number("1")
            else:
                hash_table[hash(p)] = p
                expr_table[hash(p)] = Number("1")

        cons = reduce(lambda x, y: x._ADD_(y), cons_table, Number("0"))

        tail = [
            (hash_table[h] if (c == Number("1")) else cls(T_MUL, c, hash_table[h]))
            for h, c in expr_table.items()
        ]

        if cons != Number("0"):
            if len(tail) == 0:
                return cons
            else:
                return cls(T_ADD, cons, *tail)
        else:
            if len(tail) == 0:
                return Number("0")
            elif len(tail) == 1:
                return tail[0]
            else:
                return cls(T_ADD, *tail)

    @staticmethod
    def R_NEG(cls: type, x: SatType) -> SatType:
        if x.is_expr and x.head in cls.TABLE[T_NEG]:
            return cls.TABLE[T_NEG][x.head](*x.tail)
        else:
            return x._NEG_()

    @staticmethod
    def R_MUL(cls: type, *tail: SatType):
        cons_table = []
        expr_table = []

        even = True

        for p in tail:
            if p.is_number:
                if p < 0:
                    even = not even
                    cons_table.append(abs(p))
                else:
                    cons_table.append(p)
            elif p.is_expr and p.head == T_NEG:
                even = not even
                expr_table.append(p[1])
            else:
                expr_table.append(p)

        cons = reduce(lambda x, y: x._MUL_(y), cons_table, Number("1"))

        if cons != Number("1"):
            if len(expr_table) == 0:
                return cons if even else -cons
            else:
                return (
                    cls(T_MUL, cons, *expr_table)
                    if even
                    else cls(T_MUL, -cons, *expr_table)
                )
        else:
            if len(expr_table) == 0:
                return Number("1") if even else Number("-1")
            elif len(expr_table) == 1:
                return expr_table[0] if even else cls(T_NEG, expr_table[0])
            else:
                return (
                    cls(T_MUL, *expr_table)
                    if even
                    else cls(T_MUL, cls(T_NEG, expr_table[0]), *expr_table[1:])
                )

    @staticmethod
    def R_MUL_EXPAND(cls: type, *tail: tuple):
        fulltail = []

        if not any([p.is_expr and (p.head == T_ADD) for p in tail]):
            return cls(T_MUL, *tail)

        for p in tail:
            if p.is_expr and (p.head == T_ADD):
                if fulltail:
                    fulltail = [[*x, y] for x in fulltail for y in p.tail]
                else:
                    fulltail = [[y] for y in p.tail]
            else:
                if fulltail:
                    fulltail = [[*x, p] for x in fulltail]
                else:
                    fulltail = [[p]]

        fulltail = [q[0] if (len(q) == 1) else cls(T_MUL, *q) for q in fulltail]

        if len(fulltail) == 0:
            return Number("1")
        elif len(fulltail) == 1:
            return fulltail[0]
        else:
            return cls(T_ADD, *fulltail)

    RULES = {}

    @classmethod
    def get_rules(cls) -> dict:
        return {
            ## Logical
            T_OR: [cls.R_OR],
            T_AND: [cls.R_AND],
            T_NOT: [cls.R_NOT],
            T_XOR: [lambda _, x, y: x._XOR_(y)],
            T_IMP: [lambda _, x, y: x._IMP_(y)],
            T_RIMP: [lambda _, x, y: y._IMP_(x)],
            T_IFF: [lambda _, x, y: x._IFF_(y)],
            ## Arithmetic
            T_ADD: [cls.R_ADD],
            T_NEG: [cls.R_NEG],
            T_MUL: [cls.R_MUL, cls.R_MUL_EXPAND],
            T_DIV: [lambda _, x, y: x._DIV_(y)],
            T_MOD: [lambda _, x, y: x._MOD_(y)],
            ## Comparisons
            T_EQ: [lambda _, x, y: x._EQ_(y)],
            T_NE: [lambda _, x, y: x._NE_(y)],
            T_GE: [lambda _, x, y: x._GE_(y)],
            T_LE: [lambda _, x, y: x._LE_(y)],
            T_GT: [lambda _, x, y: x._GT_(y)],
            T_LT: [lambda _, x, y: x._LT_(y)],
            ## Indexing
            T_IDX: [lambda _, x, *i: x._IDX_(i)],
        }

    @classmethod
    def apply_rule(cls, expr: SatType) -> SatType:
        """\
        Applies rules as described in `cls.RULES` for each expression, in a backward fashion i.e. from leaves to root, then, applies forward. As result, may yield Var, Number or Expr.

        Parameters
        ----------
        expr : SatType
            Input expression, variable or constant.

        Returns
        -------
        SatType
        """
        if expr.is_expr:
            if expr.head in cls.RULES:
                e = expr
                for rule in cls.RULES[expr.head]:
                    if e.is_expr and (expr.head == e.head):
                        e = rule(cls, *e.tail)
                    else:
                        return cls.apply_rule(e)
                else:
                    return e
            else:
                return expr
        else:
            return expr

    @classmethod
    def calculate(cls, expr: SatType) -> SatType:
        """\
        Applies rules as described in `cls.RULES` for each expression, in a backward fashion i.e. from leaves to root, then, applies forward. As result, may yield Var, Number or Expr.
        
        Parameters
        ----------
        expr : SatType
            Input expression, variable or constant.

        Returns
        -------
        SatType
        """
        return cls.apply(cls.back_apply(expr, cls.apply_rule), cls.apply_rule)

    @classmethod
    def transverse(cls, e: SatType, f: Callable, *args, **kwargs) -> None:
        """
        Transverses expression tree calling `func` on leaves.

        Parameters
        ----------
        e : SatType
            Expression to be transversed.
        f : Callable
            Function `f(e: SatType, *args, **kwargs)`.
        *args : tuple (optional)
            `f`'s positional arguments
        **kwargs : dict (optional)
            `f`'s key-word arguments
        """
        if e.is_expr:
            for x in e.tail:
                cls.transverse(x, f, *args, **kwargs)
        else:
            f(e, *args, **kwargs)

    @classmethod
    def seek(cls, e: SatType, f, *a, **kw) -> list:
        """\
        Seeks, depth-first, for tree leaves who satisfy `func`.

        Parameters
        ----------
        e : SatType
            Expression to be seeked.
        f : Callable
            Function `f(e: SatType, *args, **kwargs) -> bool`.
        *a : tuple (optional)
            `f`'s positional arguments
        **kw : dict (optional)
            `f`'s key-word arguments
        """
        if e.is_expr:
            return [y for x in e.tail for y in cls.seek(x, f, *a, **kw)]
        else:
            if f(e, *a, **kw):
                return [e]
            else:
                return []

    @classmethod
    def tell(cls, e: SatType, c: Callable, f: Callable, *a: tuple, **kw: dict) -> bool:
        """Returns either `True` or `False` if `choice` subtrees
        satisfies `func`.
        """
        if e.is_expr:
            h = (f(e, *a, **kw),)
            t = (cls.tell(x, c, f, *a, **kw) for x in e.tail)
            return c(it.chain(h, t))
        else:
            return f(e, *a, **kw)

    @classmethod
    def apply(cls, e: SatType, f: Callable, *a: tuple, **kw: dict) -> SatType:
        """\
            Forward-applies function `f` to `e`.

        >>> def func(expr, *args, **kwargs):
        ...    new_expr = ...
        ...    return new_expr
        """

        e = f(e, *a, **kw)

        if e.is_expr:
            t = (cls.apply(x, f, *a, **kw) for x in e.tail)
            return cls(e.head, *t)
        else:
            return e

    @classmethod
    def back_apply(cls, e: SatType, f: Callable, *a: tuple, **kw: dict):
        """Backward-applies function `func` to `expr`."""
        if e.is_expr:
            t = (cls.back_apply(x, f, *a, **kw) for x in e.tail)
            return f(cls(e.head, *t, *a, **kw))
        else:
            return f(e, *a, **kw)

    @classmethod
    def sub(cls, e: SatType, x: SatType, y: SatType) -> SatType:
        """\
        Substitutes 'x' for 'y' in 'e'.

        Parameters
        ----------
        e : SatType
            Expression to be seeked.
        f : Callable
            Function `f(e: SatType, *args, **kwargs) -> bool`.
        *a : tuple (optional)
            `f`'s positional arguments
        **kw : dict (optional)
            `f`'s key-word arguments
        """
        return cls.apply(e, lambda z: y if z == x else z)

    # -*- Subtype Checking -*-
    @property
    def is_expr(self) -> bool:
        return True

    @property
    def is_number(self) -> bool:
        return False

    @property
    def is_array(self) -> bool:
        return False

    @property
    def is_var(self) -> bool:
        return False

    # -*- Comparison Operator Definition -*-
    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    def __ne__(self, other) -> bool:
        return hash(self) != hash(other)

    # -*- Python Magic Method Aliases -*-
    def __invert__(self):
        return self._NOT_()

    def __and__(self, other):
        return self._AND_(other)

    def __or__(self, other):
        return self._OR_(other)

    def __xor__(self, other):
        return self._XOR_(other)

    def __neg__(self):
        return self._NEG_()

    def __add__(self, other):
        return self._ADD_(other)

    def __mul__(self, other):
        return self._MUL_(other)

    # -*- Other Transforms -*-
    TABLE = {
        T_NOT: {
            T_AND: lambda *X: Expr(T_OR, *(~x for x in X)),
            T_OR: lambda *X: Expr(T_AND, *(~x for x in X)),
            T_XOR: lambda A, B: A.__iff__(B),
            T_IMP: lambda A, B: A & ~B,
            T_RIMP: lambda A, B: ~A & B,
            T_IFF: lambda A, B: A ^ B,
        },
        T_IMP: {
            T_IMP: (lambda A, B: A._NOT_()._OR_(B)),
            T_RIMP: (lambda A, B: A._OR_(B._NOT_())),
            T_IFF: (lambda A, B: (A._NOT_()._AND_(B._NOT_())._OR_(A._AND_(B)))),
        },
        T_XOR: {
            T_XOR: (lambda A, B: (~A & B) | (A & ~B)),
        },
        T_NEG: {
            # AND, OR, XOR:
            T_ADD: (lambda *X: Expr(T_ADD, *(x._NEG_() for x in X))),
            T_MUL: (lambda x, *Y: Expr(T_MUL, x._NEG_(), *Y)),
            T_NEG: (lambda x: x),
            # Implications:
            T_DIV: (lambda x, y: x._NEG_()._DIV_(y)),
        },
    }

    @classmethod
    def _remove_implications(cls: type, expr: Expr) -> Expr:
        if expr.is_expr and expr.head in cls.TABLE[T_IMP]:
            return cls.TABLE[T_IMP][expr.head](*expr.tail)
        elif expr.is_expr and expr.head in cls.TABLE[T_XOR]:
            return cls.TABLE[T_XOR][expr.head](*expr.tail)
        else:
            return expr

    @classmethod
    def _move_not_inwards(cls: type, expr: Expr) -> Expr:
        if expr.is_expr and expr.head == T_NOT:
            expr = expr.tail[0]
            if expr.is_expr:
                if expr.head in cls.TABLE[T_NOT]:
                    return cls.TABLE[T_NOT][expr.head](*expr.tail)
                else:
                    return cls(T_NOT, expr)
            else:
                return cls(T_NOT, expr)
        else:
            return expr

    @classmethod
    def _distribute_and_over_or(cls: type, expr: Expr) -> Expr:
        if expr.is_expr and expr.head == T_OR:
            conjs = [
                (p.tail if (p.is_expr and p.head == T_AND) else (p,)) for p in expr.tail
            ]
            return cls(
                T_AND,
                *(
                    cls(T_OR, *z)
                    for z in reduce(
                        lambda X, Y: [(*x, y) for x in X for y in Y], conjs, [()]
                    )
                ),
            )
        else:
            return expr

    @classmethod
    def _distribute_or_over_and(cls: type, expr: Expr) -> Expr:
        if expr.is_expr and expr.head == T_AND:
            conjs = [
                (p.tail if (p.is_expr and p.head == T_OR) else (p,)) for p in expr.tail
            ]
            return cls(
                T_OR,
                *(
                    cls(T_AND, *z)
                    for z in reduce(
                        lambda X, Y: [(*x, y) for x in X for y in Y], conjs, [()]
                    )
                ),
            )
        else:
            return expr

    @classmethod
    def _cnf(cls: type, expr: Expr) -> Expr:
        expr = cls.apply(
            expr, lambda e: cls._move_not_inwards(cls._remove_implications(e))
        )
        return cls.back_apply(expr, cls._distribute_and_over_or)

    @property
    def cnf(self) -> Expr:
        return self.__class__._cnf(self)

    @classmethod
    def _dnf(cls: type, expr: Expr) -> Expr:
        expr = cls.apply(
            expr, lambda e: cls._move_not_inwards(cls._remove_implications(e))
        )
        return cls.back_apply(expr, cls._distribute_or_over_and)

    @property
    def dnf(self) -> Expr:
        return self.__class__._dnf(self)

    @classmethod
    def _move_neg_inwards(cls: type, expr: Expr) -> Expr:
        """"""
        if expr.is_expr and (expr.head == T_NEG):
            expr = expr.tail[0]
            if expr.is_expr:
                if expr.head in cls.TABLE[T_NEG]:
                    return cls.TABLE[T_NEG][expr.head](*expr.tail)
                else:
                    return cls(T_NEG, expr)
            else:
                return cls(T_NEG, expr)
        else:
            return expr

    @classmethod
    def _remove_subtractions(cls: type, expr: Expr) -> Expr:
        """"""
        if expr.is_expr and expr.head == T_NEG and len(expr.tail) == 2:
            return cls(T_ADD, expr.tail[0], cls(T_NEG, expr.tail[1]))
        else:
            return expr

    @classmethod
    def _distribute_add_over_mul(cls: type, expr: Expr) -> Expr:
        """"""
        if expr.is_expr and expr.head == T_MUL:
            conjs = [
                (p.tail if (p.is_expr and p.head == T_ADD) else (p,)) for p in expr.tail
            ]
            return cls(
                T_ADD,
                *(
                    cls(T_MUL, *z)
                    for z in reduce(
                        lambda X, Y: [(*x, y) for x in X for y in Y], conjs, [()]
                    )
                ),
            )
        else:
            return expr

    @classmethod
    def _anf(cls: type, expr: Expr) -> Expr:
        """"""
        expr = cls.apply(
            expr, lambda e: cls._move_neg_inwards(cls._remove_subtractions(e))
        )
        return cls.back_apply(expr, cls._distribute_add_over_mul)

    @property
    def anf(self) -> Expr:
        return self.__class__._anf()

    @classmethod
    def _logical(cls: type, item: SatType):
        return (not item.is_expr) or (item.head in T_LOGICAL) or (item.head in T_EXTRA)

    @property
    def logical(self) -> bool:
        return self.tell(self, all, self._logical)

    @classmethod
    def _arithmetic(cls: type, item: SatType):
        return (
            (not item.is_expr) or (item.head in T_ARITHMETIC) or (item.head in T_EXTRA)
        )

    @property
    def arithmetic(self) -> bool:
        return self.tell(self, all, self._arithmetic)


"""
"""
# Future Imports
from __future__ import annotations

# Standard Library
import itertools as it
from pathlib import Path


class Source(str):
    """This source code object aids the tracking of tokens in order to
    indicate error position on exception handling.
    """

    LEXKEYS = {"lexpos", "chrpos", "lineno", "source"}

    def __new__(
        cls,
        *,
        fname: str = None,
        buffer: str = None,
        offset: int = 0,
        length: int = None,
    ):
        """This object is a string itself with additional features for
        position tracking.
        """
        if fname is not None and buffer is not None:
            raise ValueError(
                "Can't work with both 'fname' and 'buffer' parameters, choose one option."
            )

        elif fname is not None:
            if not isinstance(fname, (str, Path)):
                raise TypeError(
                    f"Invalid type '{type(fname)}' for 'fname'. Must be 'str' or 'Path'."
                )

            fpath = Path(fname)

            if not fpath.exists() or not fpath.is_file():
                raise FileNotFoundError(f"Invalid file path '{fname}'.")

            with open(fpath, mode="r", encoding="utf-8") as file:
                return super(Source, cls).__new__(cls, file.read())

        elif buffer is not None:
            if not isinstance(buffer, str):
                raise TypeError(
                    f"Invalid type '{type(buffer)}' for 'buffer'. Must be 'str'."
                )

            return super(Source, cls).__new__(cls, buffer)
        else:
            raise ValueError("Either 'fname' or 'buffer' must be provided.")

    def __init__(
        self,
        *,
        fname: str = None,
        buffer: str = None,
        offset: int = 0,
        length: int = None,
    ):
        """Separates the source code in multiple lines. A blank first line is added for the indexing to start at 1 instead of 0. `self.table` keeps track of the (cumulative) character count."""
        if not isinstance(offset, int) or offset < 0:
            raise TypeError("'offset' must be a positive integer (int).")
        elif length is None:
            length = len(self)
        elif not isinstance(length, int) or length < 0:
            raise TypeError("'length' must be a positive integer (int) or 'None'.")

        self.offset = min(offset, len(self))
        self.length = min(length, len(self) - self.offset)

        self.fpath = (
            Path(fname).resolve(strict=True) if (fname is not None) else "<string>"
        )
        self.lines = [""] + self.split("\n")
        self.table = list(it.accumulate([(len(line) + 1) for line in self.lines]))

    def __str__(self):
        return self[self.offset : self.offset + self.length]

    def __repr__(self):
        return f"{self.__class__.__name__}({self.fpath!r})"

    def __bool__(self):
        """Truth-value for emptiness checking."""
        return self.__len__() > 0

    def getlex(self, lexpos: int = None) -> dict:
        """Retrieves lexinfo dictionary from lexpos."""
        if lexpos is None:
            return self.eof.lexinfo
        elif not isinstance(lexpos, int):
            raise TypeError(f"'lexpos' must be an integer (int), not ({type(lexpos)}).")
        elif not 0 <= lexpos <= self.length:
            return self.eof.lexinfo

        lexpos = lexpos + self.offset + 1

        lineno = 1
        while lineno < len(self.table) and lexpos >= self.table[lineno]:
            lineno += 1

        if lineno == len(self.table):
            return self.eof.lexinfo
        else:
            return {
                "lineno": lineno,
                "lexpos": lexpos,
                "chrpos": lexpos - self.table[lineno - 1],
                "source": self,
            }

    def slice(self, offset: int = 0, length: int = None):
        return self.__class__(fname=self.fpath, offset=offset, length=length)

    def error(self, msg: str, *, target: object = None, name: str = None):
        if target is None or not self.trackable(target):
            if name is not None:
                return f"In '{self.fpath}':\n" f"{name}: {msg}\n"
            else:
                return f"In '{self.fpath}':\n" f"{msg}\n"
        else:
            if name is not None:
                return (
                    f"In '{self.fpath}' at line {target.lineno}:\n"
                    f"{self.lines[target.lineno]}\n"
                    f"{' ' * target.chrpos}^\n"
                    f"{name}: {msg}\n"
                )
            else:
                return (
                    f"In '{self.fpath}' at line {target.lineno}:\n"
                    f"{self.lines[target.lineno]}\n"
                    f"{' ' * target.chrpos}^\n"
                    f"{msg}\n"
                )

    class EOF(object):
        pass

    @property
    def eof(self):
        """Virtual object to represent the End-of-File for the given source
        object. It's an anonymously created EOFType instance.
        """
        eof = self.EOF()

        self.track(eof, len(self))

        return eof

    # -*- Tracking -*-
    def track(self, o: object, lexpos: int = None):
        """"""
        setattr(o, "lexinfo", self.getlex(lexpos))

        if not hasattr(o.__class__, "__lextrack__"):
            setattr(
                o.__class__, "chrpos", property(lambda this: this.lexinfo["chrpos"])
            )
            setattr(
                o.__class__, "lineno", property(lambda this: this.lexinfo["lineno"])
            )
            setattr(
                o.__class__, "lexpos", property(lambda this: this.lexinfo["lexpos"])
            )
            setattr(
                o.__class__, "source", property(lambda this: this.lexinfo["source"])
            )
            setattr(o.__class__, "__lextrack__", None)

    @classmethod
    def blank(cls, o: object):
        setattr(o, "lexinfo", {"chrpos": 0, "lineno": 0, "lexpos": 0, "source": None})

        if not hasattr(o.__class__, "__lextrack__"):
            setattr(
                o.__class__, "chrpos", property(lambda this: this.lexinfo["chrpos"])
            )
            setattr(
                o.__class__, "lineno", property(lambda this: this.lexinfo["lineno"])
            )
            setattr(
                o.__class__, "lexpos", property(lambda this: this.lexinfo["lexpos"])
            )
            setattr(
                o.__class__, "source", property(lambda this: this.lexinfo["source"])
            )
            setattr(o.__class__, "__lextrack__", None)

    def propagate(self, x: object, y: object, *, out: bool = False) -> object | None:
        if self.trackable(x, strict=True) and self.trackable(y):
            y.lexinfo.update(x.lexinfo)
            if out:
                return y
            else:
                return None
        else:
            raise TypeError(
                f"Can't propagate lexinfo between types {type(x)} and {type(y)}"
            )

    @classmethod
    def trackable(cls, o: object, *, strict: bool = False):
        if cls._trackable(o):
            return True
        elif strict:
            print(o, o.lexinfo)
            raise TypeError(f"Object '{o}' of type '{type(o)}' is not trackable.")
        else:
            return False

    @classmethod
    def _trackable(cls, o: object):
        if not hasattr(o, "lexinfo") or not isinstance(o.lexinfo, dict):
            return False
        else:
            if any(key not in o.lexinfo for key in cls.LEXKEYS):
                return False
            else:
                if (
                    not hasattr(o, "lineno")
                    or not isinstance(o.lineno, int)
                    or o.lineno < 0
                ):
                    return False
                elif (
                    not hasattr(o, "lexpos")
                    or not isinstance(o.lexpos, int)
                    or o.lexpos < 0
                ):
                    return False
                elif (
                    not hasattr(o, "chrpos")
                    or not isinstance(o.chrpos, int)
                    or o.chrpos < 0
                ):
                    return False
                elif (
                    not hasattr(o, "source")
                    or not isinstance(o.source, Source)
                    and not o.source is None
                ):
                    return False
                else:
                    return True


__all__ = ["Source"]
